;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  (__)  LAMBIC Project
;   )(   Written by Jorge Vallejos
;        
; (c) Vrije Universiteit Brussel, 2008 - 2010
; 
; Permission is hereby granted, free of charge, to any person
; obtaining a copy of this software and associated documentation
; files (the "Software"), to deal in the Software without
; restriction, including without limitation the rights to use,
; copy, modify, merge, publish, distribute, sublicense, and/or
; sell copies of the Software, and to permit persons to whom the
; Software is furnished to do so, subject to the following
; conditions:
;
; The above copyright notice and this permission notice shall be
; included in all copies or substantial portions of the Software.
;
; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
; EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
; OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
; NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
; HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
; WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
; OTHER DEALINGS IN THE SOFTWARE.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; LAMBIC GUEZE Drawing Editor
; CAPI class extensions
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(in-package :lambic-user)

; Global parameters and variables
(defvar *brush-cursor* 
      (capi:load-cursor
       '((:cocoa
          #.(cl-user::current-pathname "images/brush-cursor.tif")
          :x-hot 2
          :y-hot 40))))

(defvar *br-finger-cursor* 
      (capi:load-cursor
       '((:cocoa
          #.(cl-user::current-pathname "images/br-finger-cursor.tif")
          :x-hot 17
          :y-hot 11))))

(defvar *tl-finger-cursor*
  (make-instance 'capi:image-pinboard-object
                 :image #.(cl-user::current-pathname "images/tl-finger-cursor.tif")))

(color:define-color-alias :shadow-color
                          (color:make-rgb 0.15s0 0.51s0 0.93s0 0.4s0))


(color:define-color-alias :selection-color
                           (color:make-rgb 0.15s0 0.51s0 0.93s0 0.4s0))


(defvar *editor-index* 0)

(defvar *shadow-colors* (list (color:make-rgb 0.51s0 0.93s0 0.15s0 0.4s0)
                              (color:make-rgb 0.15s0 0.51s0 0.93s0 0.4s0)
                              (color:make-rgb 0.93s0 0.15s0 0.93s0 0.4s0)
                              (color:make-rgb 0.93s0 0.51s0 0.51s0 0.4s0)
                              (color:make-rgb 0.51s0 0.93s0 0.93s0 0.4s0)))

; FIGURE-STATUS struct
; Used to keep the status of an figure while it is being drawn, resized or moved.
(defstruct figure-status x y width height)

; DRAG-STATUS struct
(defstruct drag-status x y)

;;;;;;;;;;
; EDITOR ;
;;;;;;;;;;

; The gueze drawing editor class.
(defgroupclass gueze-editor (capi:pinboard-layout)
  ((username :initarg :username :accessor editor-username)
   (figures :initform nil :accessor editor-figures)
   (canvas :accessor editor-canvas)
   (brush-active :initform nil :accessor brush-activep)
   (resize-active :initform nil :accessor resize-activep)
   (selected-color :initform :black :accessor selected-color)
   (zoom-level :initform 1 :accessor zoom-level)
   ; temporal fields
   (drag-status :initform nil :accessor drag-status)
   (current-line :initform nil :accessor current-line)
   (current-point :initform nil :accessor current-point)
   (selection-rectangle :initform nil :accessor selection-rectangle)
   ; collaboration fields
   (session :initform nil :initarg :session :accessor in-session)
   (index :initform -1 :accessor editor-index)
   (leaderp :initform nil :accessor is-leader)))

; Initialise an GUEZE-EDITOR instance by adding the canvas.
(c2cl:defmethod initialize-instance :after ((this-editor gueze-editor)
                                            &key &allow-other-keys)
  (let ((canvas (make-instance 'capi:rectangle
                               :width 700
                               :height 500
                               :graphics-args '(:foreground :white)
                               :filled t)))
    (capi:manipulate-pinboard this-editor canvas :add-top)
    (setf (editor-canvas this-editor) canvas)))

;;;;;;;;;;
; FIGURE ;
;;;;;;;;;;

; The Gueze figure class
(defgroupclass gueze-figure (capi:pinboard-object)
  ((name :initarg :name :accessor figure-name)
   (owner :initarg :owner :accessor figure-owner)
   ; collaboration fields
   (user :initarg :user :initform nil :accessor figure-user)
   (version :initform 0 :accessor figure-version)))

; Rectangle
(defclass gueze-rectangle (gueze-figure) ())

; Ellipse
(defclass gueze-ellipse (gueze-figure) ())

; Drawn line
(defclass gueze-drawn-figure (gueze-figure)
  ((line :initform nil :initarg :line :accessor line)
   (status :initform nil :accessor status)))

; Get figure at a given position.
(defun get-figure-at-position (editor x y)
  (let ((figure (capi:pinboard-object-at-position editor x y)))
    (unless (eql (editor-canvas editor) figure)
      figure)))

; Generate a figure name
(defun generate-figure-name (name)
  (read-from-string (format nil "~d-fig-~d" name (gensym))))

(defmethod find-selection-color (figure editor)
  (if (< (editor-index editor) 0)
      :shadow-color
    (let ((user (figure-user figure)))
      (nth (peer-position user (group-peers editor))
           *shadow-colors*))))

#|
  (if (> (editor-index editor) 0)
      :shadow-color
    (let ((user (figure-user figure)))
      (nth (peer-position user (group-peers editor))
           *shadow-colors*))))
|#
#|
  (if (peer-equal user editor)
      :shadow-color
    (nth (position user (group-peers editor))
         *shadow-colors*)))
|#
  
; CAPI method
; Draw Gueze rectangle
(c2cl:defmethod capi:draw-pinboard-object ((editor gueze-editor) (rectangle gueze-rectangle)
                                           &key &allow-other-keys)
  (capi:with-geometry rectangle
    (let* ((args (capi:pinboard-object-graphics-args rectangle))
           (level (zoom-level editor))
           (position-offset (* 4 level))
           (size-offset (* 8 level)))
      (when (figure-user rectangle)
        (gp:draw-rectangle editor capi:%x% capi:%y% capi:%width% capi:%height%
                           :foreground (find-selection-color rectangle editor)
                           :filled t))
      (gp:draw-rectangle editor (+ capi:%x% position-offset) (+ capi:%y% position-offset)
                         (- capi:%width% size-offset) (- capi:%height% size-offset)
                         :foreground (getf args :foreground) :filled t))))

; CAPI method
; Draw Gueze ellipse
(c2cl:defmethod capi:draw-pinboard-object ((editor gueze-editor) (ellipse gueze-ellipse)
                                      &key &allow-other-keys)
  (capi:with-geometry ellipse
    (let* ((args (capi:pinboard-object-graphics-args ellipse))
           (half-width  (floor (1- capi:%width%)  2))
           (half-height (floor (1- capi:%height%) 2))
           (circle-x (+ capi:%x% half-width))
           (circle-y (+ capi:%y% half-height))
           (level (zoom-level editor))
           (size-offset (* 4 level)))
      (when (figure-user ellipse)
        (gp:draw-ellipse editor circle-x circle-y half-width half-height 
                         :foreground (find-selection-color ellipse editor)
                         :filled t))
      (gp:draw-ellipse editor circle-x circle-y
                       (- half-width size-offset) (- half-height size-offset)
                       :foreground (getf args :foreground) :filled t))))

; Move line
(defun move-line (line status new-x new-y)
  (with-slots (x y) status
      (let ((delta-x (- new-x x))
            (delta-y (- new-y y)))
        (unless (and (zerop delta-x) (zerop delta-y))
          (loop for element in line
                for counter from 0 to (1- (length line))
                when (evenp counter) collect (+ element delta-x)
                else collect (+ element delta-y))))))

; Resize line
(defun resize-line (line status new-width new-height)
  (with-slots (x y width height) status
      (let ((delta-w (/ new-width width))
            (delta-h (/ new-height height)))
        (unless (and (eql delta-w 1) (eql delta-h 1))
          (loop for element in line
                for counter from 0 to (1- (length line))
                when (evenp counter)
                collect (+ x (* (- element x) delta-w))
                else
                collect (+ y (* (- element y) delta-h)))))))

; CAPI method
; Set figure's size
(c2cl:defmethod capi::set-pinboard-pane-size ((figure gueze-drawn-figure) width height)
  (let ((status (status figure)))
    (setf (line figure)
          (resize-line (line figure) status width height))
    (call-next-method)
    (setf (figure-status-width status) width)
    (setf (figure-status-height status) height)))

; CAPI method
; Set figure's position
(c2cl:defmethod capi::set-pinboard-pane-position ((figure gueze-drawn-figure) x y)
  (let ((status (status figure)))
    (setf (line figure)
          (move-line (line figure) status x y))
    (call-next-method)
    (setf (figure-status-x status) x)
    (setf (figure-status-y status) y)))

; CAPI method
; Set figure's size and position
(c2cl:defmethod set-pinboard-object-geometry ((figure gueze-drawn-figure)
                                           &key x y width height)
  (let* ((status (status figure))
         (line (line figure))
         (delta-w (/ width (figure-status-width status)))
         (delta-h (/ height (figure-status-height status))))
    (unless (and (eql delta-w 1) (eql delta-h 1))
      (setf (line figure)
            (loop for element in line
                for counter from 0 to (1- (length line))
                when (evenp counter) collect (* element delta-w)
                else collect (* element delta-h)))
      (call-next-method)
      (setf (status figure) (make-figure-status :x x :y y
                                            :width width :height height)))))

; CAPI method
; Draw Gueze drawn figure
(c2cl:defmethod capi:draw-pinboard-object ((editor gueze-editor) (figure gueze-drawn-figure)
                                      &key &allow-other-keys)
  (let* ((line (line figure))
         (args (capi:pinboard-object-graphics-args figure))
         (level (zoom-level editor))
         (shadow-thickness (* 10 level))
         (line-thickness (* 4 level)))
    (when (figure-user figure)
      (gp:draw-lines editor line
                     :foreground (find-selection-color figure editor)
                     :thickness shadow-thickness
                     :line-end-style :round :line-joint-style :round))
    (gp:draw-lines editor line :foreground (getf args :foreground)
                   :line-end-style :round :line-joint-style :round :thickness line-thickness)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;
; CACHE FUNCTIONS ;
;;;;;;;;;;;;;;;;;;;
; To be done automatically

; CACHE FUNCTION for EDITOR-INDEX/(SETF EDITOR-INDEX)
(defmethod editor-index ((remote-object generic-channel))
  (let ((cache-future (find-cache-function-future remote-object
                                                  :observer-function (fdefinition 'editor-index)
                                                  :provider-function (fdefinition '(setf editor-index)))))
    (if (resolvedp cache-future)
        (future-result cache-future)
      (progn 
        (resolve-with-result cache-future
                             (in-actor-of remote-object (editor-index remote-object)))
        cache-future))))

(defmethod editor-index ((reference far-reference))
  (editor-index (provide-channel *communication-actor* reference)))

(defmethod (setf editor-index) :after (value (local-object gueze-editor))
  (let ((futures (find-cache-function-futures local-object (fdefinition '(setf editor-index)))))
    (loop for future in futures
          do (in-actor-of future (resolve-with-result future value)))))

; CACHE FUNCTION for IS-LEADER/(SETF IS-LEADER)
(defmethod is-leader ((remote-object generic-channel))
  (let ((cache-future (find-cache-function-future remote-object
                                                  :observer-function (fdefinition 'is-leader)
                                                  :provider-function (fdefinition '(setf is-leader)))))
    (if (resolvedp cache-future)
        (future-result cache-future)
      (progn 
        (resolve-with-result cache-future
                             (in-actor-of remote-object (is-leader remote-object)))
        cache-future))))

(defmethod is-leader ((reference far-reference))
  (is-leader (provide-channel *communication-actor* reference)))

(defmethod (setf is-leader) :after (value (local-object gueze-editor))
  (let ((futures (find-cache-function-futures local-object (fdefinition '(setf is-leader)))))
    (loop for future in futures
          do (in-actor-of future (resolve-with-result future value)))))

; CACHE FUNCTION for EDITOR-USERNAME/(SETF EDITOR-USERNAME)
(defmethod editor-username ((remote-object generic-channel))
  (let ((cache-future (find-cache-function-future remote-object
                                                  :observer-function (fdefinition 'editor-username)
                                                  :provider-function (fdefinition '(setf editor-username)))))
    (if (resolvedp cache-future)
        (future-result cache-future)
      (progn 
        (resolve-with-result cache-future
                             (in-actor-of remote-object (editor-username remote-object)))
        cache-future))))

(defmethod editor-username ((reference far-reference))
  (editor-username (provide-channel *communication-actor* reference)))

(defmethod (setf editor-username) :after (value (local-object gueze-editor))
  (let ((futures (find-cache-function-futures local-object (fdefinition '(setf editor-username)))))
    (loop for future in futures
          do (in-actor-of future (resolve-with-result future value)))))

; CACHE FUNCTION for IN-SESSION/(SETF IN-SESSION)
(defmethod in-session ((remote-object generic-channel))
  (let ((cache-future (find-cache-function-future remote-object
                                                  :observer-function (fdefinition 'in-session)
                                                  :provider-function (fdefinition '(setf in-session)))))
    (if (resolvedp cache-future)
        (future-result cache-future)
      (progn 
        (resolve-with-result cache-future
                             (in-actor-of remote-object (in-session remote-object)))
        cache-future))))

(defmethod in-session ((reference far-reference))
  (in-session (provide-channel *communication-actor* reference)))

(defmethod (setf in-session) :after (value (local-object gueze-editor))
  (let ((futures (find-cache-function-futures local-object (fdefinition '(setf in-session)))))
    (loop for future in futures
          do (in-actor-of future (resolve-with-result future value)))))

; CACHE FUNCTION for FIGURE-NAME/(SETF FIGURE-NAME)
(defmethod figure-name ((remote-object generic-channel))
  (let ((cache-future (find-cache-function-future remote-object
                                                  :observer-function (fdefinition 'figure-name)
                                                  :provider-function (fdefinition '(setf figure-name)))))
    (if (resolvedp cache-future)
        (future-result cache-future)
      (progn 
        (resolve-with-result cache-future
                             (in-actor-of remote-object (figure-name remote-object)))
        cache-future))))

(defmethod figure-name ((reference far-reference))
  (figure-name (provide-channel *communication-actor* reference)))

(defmethod (setf figure-name) :after (value (local-object gueze-figure))
  (let ((futures (find-cache-function-futures local-object (fdefinition '(setf figure-name)))))
    (loop for future in futures
          do (in-actor-of future (resolve-with-result future value)))))

; CACHE FUNCTION for FIGURE-OWNER/(SETF FIGURE-OWNER)
(defmethod figure-owner ((remote-object generic-channel))
  (let ((cache-future (find-cache-function-future remote-object
                                                  :observer-function (fdefinition 'figure-owner)
                                                  :provider-function (fdefinition '(setf figure-owner)))))
    (if (resolvedp cache-future)
        (future-result cache-future)
      (progn 
        (resolve-with-result cache-future
                             (in-actor-of remote-object (figure-owner remote-object)))
        cache-future))))

(defmethod figure-owner ((reference far-reference))
  (figure-owner (provide-channel *communication-actor* reference)))
  
(defmethod (setf figure-owner) :after (value (local-object gueze-figure))
  (let ((futures (find-cache-function-futures local-object (fdefinition '(setf figure-owner)))))
    (loop for future in futures
          do (in-actor-of future (resolve-with-result future value)))))

; CACHE FUNCTION for FIGURE-USER/(SETF FIGURE-USER)
(defmethod figure-user ((remote-object generic-channel))
  (let ((cache-future (find-cache-function-future remote-object
                                                  :observer-function (fdefinition 'figure-user)
                                                  :provider-function (fdefinition '(setf figure-user)))))
    (if (resolvedp cache-future)
        (future-result cache-future)
      (progn 
        (resolve-with-result cache-future
                             (in-actor-of remote-object (figure-user remote-object)))
        cache-future))))

(defmethod figure-user ((reference far-reference))
  (figure-user (provide-channel *communication-actor* reference)))

(defmethod (setf figure-user) :after (value (local-object gueze-figure))
  (let ((futures (find-cache-function-futures local-object (fdefinition '(setf figure-user)))))
    (loop for future in futures
          do (in-actor-of future (resolve-with-result future value)))))

; CACHE FUNCTION for FIGURE-VERSION/(SETF FIGURE-VERSION)
(defmethod figure-version ((remote-object generic-channel))
  (let ((cache-future (find-cache-function-future remote-object
                                                  :observer-function (fdefinition 'figure-version)
                                                  :provider-function (fdefinition '(setf figure-version)))))
    (if (resolvedp cache-future)
        (future-result cache-future)
      (progn 
        (resolve-with-result cache-future
                             (in-actor-of remote-object (figure-version remote-object)))
        cache-future))))

(defmethod figure-version ((reference far-reference))
  (figure-version (provide-channel *communication-actor* reference)))

(defmethod (setf figure-version) :after (value (local-object gueze-figure))
  (let ((futures (find-cache-function-futures local-object (fdefinition '(setf figure-version)))))
    (loop for future in futures
          do (in-actor-of future (resolve-with-result future value)))))
