;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  (__)  LAMBIC Project
;   )(   Written by Jorge Vallejos
;        
; (c) Vrije Universiteit Brussel, 2008 - 2010
; 
; Permission is hereby granted, free of charge, to any person
; obtaining a copy of this software and associated documentation
; files (the "Software"), to deal in the Software without
; restriction, including without limitation the rights to use,
; copy, modify, merge, publish, distribute, sublicense, and/or
; sell copies of the Software, and to permit persons to whom the
; Software is furnished to do so, subject to the following
; conditions:
;
; The above copyright notice and this permission notice shall be
; included in all copies or substantial portions of the Software.
;
; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
; EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
; OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
; NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
; HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
; WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
; OTHER DEALINGS IN THE SOFTWARE.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; LAMBIC ACTORS
; 
; - Actors as object containers
; - Asynchronous remote method invocations
; - Asynchronous result handling (futures)
; - Message passing is only at the actor level.
;
; To test execute:
;
; (TEST-ACTOR)
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package :lambic)

; Lambic internal actors. To be initialised in the channels.lisp
; and bonjour.lisp files respectively.
(defvar *communication-actor*)
(defvar *discovery-actor*)

; Representation of a function call as struct for serialisation
(defstruct function-call name arguments)

; Representation of an actor message as struct for serialisation
(defstruct actor-message sender receiver function-call future timeout discard-value)

; Actor ID generation
(defvar *actor-id* 0)
(defun generate-new-actor-name ()
  (format nil "~d" (incf *actor-id*)))

; IN-ACTOR-OF macro
(defmacro in-actor-of (receiver (name &rest arguments)
                                &key (with-future nil with-future-supplied-p)
                                (due-in nil due-in-supplied-p)
                                (discard-with-value nil discard-with-value-p))
  `(send-to-actor-of ,receiver ',name (list ,@arguments)
                     ,@(when with-future-supplied-p `(:with-future ,with-future))
                     ,@(when due-in-supplied-p `(:due-in ,due-in))
                     ,@(when discard-with-value-p
                         `(:discard-with-value ,discard-with-value))))

; Multiple asynchronous invocation
; IN-ACTORS-OF macro
(defmacro in-actors-of ((receiver-name receivers) (name &rest arguments)
                                 &key (with-future nil with-future-supplied-p)
                                 (due-in nil due-in-supplied-p)
                                 (discard-with-value nil discard-with-value-p))
  `(map 'list 
        #'(lambda (,receiver-name)
            (send-to-actor-of ,receiver-name ',name (list ,@arguments)
                              ,@(when with-future-supplied-p
                                  `(:with-future ,with-future))
                              ,@(when due-in-supplied-p
                                  `(:due-in ,due-in))
                              ,@(when discard-with-value-p
                                  `(:discard-with-value ,discard-with-value))))
        ,receivers))

; Single asynchronous invocation
; Invoke a function at the receiver's actor
(defmethod send-to-actor-of (receiver function-name function-arguments
                             &key
                             (with-future (make-instance 'future))
                             due-in discard-with-value)
  (let ((function-call (make-function-call :name function-name
                                           :arguments function-arguments)))
    (send-actor-message *current-actor* receiver function-call with-future
                        :due-in due-in :discard-with-value discard-with-value)))

;;;;;;;;;;;;;;;
; ACTOR CLASS ;
;;;;;;;;;;;;;;;
(define-distributable-class actor ()
  ((actor-process :initarg :process :accessor actor-process)
   (actor-heart :initform t :accessor actor-heart)))

; (ASYNCHRONOUS) ACTOR MESSAGE SENDING

; CASE 1: Receiver at the same actor.
; Message is immediately received by the actor for synchronus execution.
(defmethod send-actor-message ((this actor) (receiver distributable-object) function-call future
                               &key &allow-other-keys)
  (let* ((receiver-actor (object-actor receiver))
         (message (make-actor-message :sender this :receiver receiver-actor
                                      :function-call function-call :future future)))
    (receive-actor-message receiver-actor message)
    future))

; CASE 2: Receiver at a remote actor
; Message is sent to remote actor via the communicator actor.
(defmethod send-actor-message ((this actor) (receiver far-reference) function-call future
                               &key due-in discard-with-value)
  (let* ((message (make-actor-message :sender this :receiver receiver
                                      :function-call function-call :future future
                                      :timeout due-in :discard-value discard-with-value)))
    (in-actor-of *communication-actor*
                 (send-remote-actor-message *communication-actor* message)
                 :with-future nil)
    future))

; CASE 3: Receiver is a channel
; See channels.lisp file.

; ACTOR MESSAGE RECEPTION
(defmethod receive-actor-message ((this actor) message)
  (mp:mailbox-send (mp:process-mailbox (actor-process this)) message))

; ACTOR MESSAGE PROCESSING
(defmethod process-actor-message ((this actor) message)
  (let* ((call (actor-message-function-call message))
         (future-reference (actor-message-future message)))
    (handler-case
        (let ((result (apply (function-call-name call)
                             (function-call-arguments call))))
          (when future-reference
            (if (typep result 'future)
                (when-resolved result
                  (lambda (resolved-result)
                    (in-actor-of future-reference
                                 (resolve-with-result future-reference resolved-result)
                                 :with-future nil)))
              (in-actor-of future-reference
                                 (resolve-with-result future-reference result)
                                 :with-future nil))))
      (error (e)
        (progn
#|
          (when future-reference 
            (in-actor-of future-reference
                         (ruin-with-condition future-reference
                                              (format nil "~A" e))
                         :with-future nil))
|#
          (error e))))))

; Spawn an actor, eventually including objects passed as parameters.
; These objects are said to be owned by the actor.
(defun spawn-actor (name &rest objects)
  (let ((actor (make-instance 'actor
                              :object-name (format nil "Lambic-~D-actor-at-~D"
                                                   name (machine-instance))))
        (mailbox (mp:make-mailbox)))
    (setf (object-actor actor) actor)
    (dolist (object objects) (setf (object-actor object) actor))
    (setf (actor-process actor)
          (mp:process-run-function
           (object-name actor)
           (list :mailbox mailbox)
           (lambda ()
             (let ((*current-actor* actor))
               (setf (actor-process *current-actor*) mp:*current-process*)
               (loop until (not (slot-boundp *current-actor* 'actor-heart))
                     do
                     (let ((message (mp:mailbox-read mailbox)))
                       (process-actor-message *current-actor* message)))))))
    actor))

; Kill an actor
(defun kill-actor (actor)
  (in-actor-of actor (slot-makunbound actor 'actor-heart) :with-future nil))

; Check whether the actor of the object is the *CURRENT-ACTOR*
(defmethod in-current-actor ((object distributable-object))
  (eql mp:*current-process* (actor-process *current-actor*)))
;  (eql (object-actor object) *current-actor*))

; Return NIL for far references.
(defmethod in-current-actor ((object far-reference))
  nil)

; Return T for any other kind of object.
(defmethod in-current-actor (any-other-kind-of-object)
  t)

; Initialise *current-actor*
(setf *current-actor* (spawn-actor "default"))

(defmethod actor-of ((object distributable-object))
  (object-actor object))

(defmethod actor-of ((object far-reference))
  (reference-actor-name object))

(defmethod actor-of (object)
  *current-actor*)

;;;;;;;;;;;;;;;;;;;
; Running Example ;
;;;;;;;;;;;;;;;;;;;

#|

(define-distributable-class im ()
  ((username :initarg :username :accessor username)))

(defmethod receive-text (this sender text)
  (when-resolved (in-actor-of sender (username sender))
                 (lambda (sendername)
                   (format *screen-output* "~A says: ~A" sendername text))))

(defmethod send-text (this receiver text)
  (in-actor-of receiver (receive-text receiver this text) :with-future nil))

(defvar *my-im*)
(defvar *buddy-im*)

(defun test-actor ()
  (setf *my-im* (make-instance 'im :username "Bob"))
  (setf *buddy-im* (make-instance 'im :username "Alice"))
  (spawn-actor "im" *my-im*)
  (spawn-actor "buddy-im" *buddy-im*))

(test-actor)

#|
(send-text *my-im* *buddy-im* "hello")
|#


|#