;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  (__)  LAMBIC Project
;   )(   Written by Jorge Vallejos and Pascal Costanza
;        
; (c) Vrije Universiteit Brussel, 2008 - 2010
; 
; Permission is hereby granted, free of charge, to any person
; obtaining a copy of this software and associated documentation
; files (the "Software"), to deal in the Software without
; restriction, including without limitation the rights to use,
; copy, modify, merge, publish, distribute, sublicense, and/or
; sell copies of the Software, and to permit persons to whom the
; Software is furnished to do so, subject to the following
; conditions:
;
; The above copyright notice and this permission notice shall be
; included in all copies or substantial portions of the Software.
;
; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
; EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
; OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
; NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
; HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
; WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
; OTHER DEALINGS IN THE SOFTWARE.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package :lambic)

; Variant to Common Lisp remove-if function that returns not only the
; list without the removed value, but also the removed values themselves.
(defun remove-if* (condition-function list &key (return-removed-items nil))
  (let* ((removed-elements nil)
         (new-list (remove-if #'(lambda (element)
                                     (when (funcall condition-function element)
                                       (push element removed-elements))
                                     (funcall condition-function element)) list)))
    (if return-removed-items
        (values new-list removed-elements)
      new-list)))

; Find the class that corresponds to each name.
(defun find-classes (names)
  (loop for name in names
        collect (find-class name)))

; Find the name of each class.
(defun find-names (classes)
  (loop for class in classes
        collect (class-name class)))

; Find superclass
(defun find-superclass (required-class superclasses)
  (if (not required-class)
      (error "First argument cannot be nil")
    (when (and required-class superclasses)
      (let ((superclass (car superclasses)))
        (or (eql superclass required-class)
            (if (eql superclass (find-class 'standard-object))
                (find-superclass required-class (cdr superclasses))
              (or (find-superclass required-class
                                   (class-direct-superclasses superclass))
                  (find-superclass required-class (cdr superclasses)))))))))

; Add superclass
(defun add-superclass (new-superclass class)
  (let ((superclasses (class-direct-superclasses class)))
    (unless (find-superclass new-superclass superclasses)
      (setf (class-direct-superclasses class)
            (push new-superclass superclasses)))))

; Get the value from a record that correspond to the key
; passed as parameter.
(defun get-record-value (txt-record key)
  (let ((selected-cons (find-if #'(lambda (cons) (equal (car cons) key))
                                 txt-record)))
    (cdr selected-cons)))

; Parse service description. Return the Lambic-record (object-name + actor-name)
; and the serice-record (user defined).
(defun parse-service-description (encrypted-record)
  (let ((service-description (loop
                              for encrypted-cons
                              in (dns-sd:parse-txt-record encrypted-record)
                              collect
                              (cons (car encrypted-cons)
                                    (dns-sd::bytes-to-string (cdr encrypted-cons))))))
    (multiple-value-bind (description-record lambic-record)
        (remove-if* #'(lambda (cons)
                        (or (equal (car cons) "object-name")
                            (equal (car cons) "actor-name")
                            (equal (car cons) "group-classp")))
                    service-description :return-removed-items t)
      (values description-record lambic-record))))

; Check equality of records
; At the moment this is only based on the EQUAL function.
; This is clearly something to be enriched.
(defmethod description-record-equal (required-record discovered-record)
  (loop for required-cons in required-record
        for discovered-cons in discovered-record
        when (and (equal (car required-cons) (car discovered-cons))
                  (equal (cdr required-cons) (cdr discovered-cons)))
        return t))

; Parse parameters of defmethod macro as name, qualifiers,
; specialised-lambda-list, declarations, documentation, and body.
; Additionally return argument-names which is used to define
; generic function of the method. For more details, see redefinition
; of defmethod macro in package lambic-ff.
; TODO: This does not work for methods without arguments.
; For the moment a function should be defined instead.
(defun parse-method-definition (forms)
  (let ((name (car forms)))
    (setf forms (cdr forms))
    (multiple-value-bind (qualifiers forms)
        (get-method-qualifiers forms)
      (let* ((specialised-lambda-list (car forms))
             (argument-names (get-method-argument-names specialised-lambda-list))
             (specialisers (get-method-specialisers specialised-lambda-list)))
        (setf forms (cdr forms))
        (multiple-value-bind (declarations forms)
            (get-method-declarations forms)
          (multiple-value-bind (documentation body)
              (get-method-documentation forms)
            (values name qualifiers specialised-lambda-list declarations 
                    documentation body argument-names specialisers)))))))

; Parse lambda as name, lambda-list, declarations, documentation
; and body. Body is enclosed in a futurised progn.
(defun parse-lambda (forms)
  (let ((lambda-list (first forms)))
    (multiple-value-bind (declarations forms)
        (get-method-declarations (cdr forms))
      (multiple-value-bind (documentation body)
          (get-method-documentation forms)
        (values lambda-list declarations documentation body)))))


; Get the qualifiers specified in a method definition.
; I.e., check for symbols before the method definition's lambda list.
(defun get-method-qualifiers (forms)
  (loop with newforms = forms
        for form = (pop newforms)
        if (symbolp form)
        collect form into qualifiers
        else return (values qualifiers (push form newforms))))

; Get the declarations specified in a method definition.
; I.e., check for all the forms, after the method definition's
; lambda list, starting with DECLARE.
(defun get-method-declarations (forms)
  (loop with newforms = forms
        for form = (pop newforms)
        if (and (listp form)
                (eql (car form) 'DECLARE))
        collect form into declarations
        else return (values declarations (push form newforms))))

; Get the documentation specified in a method definition.
; I.e., check for string forms after the method definition's
; lambda list.
(defun get-method-documentation (forms)
  (loop with newforms = forms
        for form = (pop newforms)
        if (stringp form)
        collect form into declarations
        else return (values declarations (push form newforms))))

; Get the argument names of the lambda list of a method
; definition.
(defun get-method-argument-names (lambda-list)
  (loop for argument in lambda-list
        if (listp argument)
        collect (first argument)
        else
        collect argument))

; Get the method specialisers' names.
(defun get-method-specialisers (lambda-list)
  (loop for argument in lambda-list
        if (listp argument)
        if (listp (second argument))
        collect (second argument)
        else
        collect (find-class (second argument))
        else
        collect (find-class t)))

; Shadow multiple functions
(defun shadow-symbols (function-names)
  (loop for function-name in function-names
        do (shadow function-name)))

; From Peter Norvig in othello.lisp
(defun mappend (fn list)
  "Append the results of calling fn on each element of list.
  Like mapcon, but uses append instead of nconc."
  (apply #'append (mapcar fn list)))

; From Pascal Costanza's implementation of filtered dispatch
(defmacro with-next-methods (next-methods &body body)
  `(let ((clos::*next-methods* ,next-methods))
     ,@body))

; Compute applicable methods after evaluating predicates
(defmethod apply-selected-methods (selected-methods function args)
  (loop for method in selected-methods
        for (qualifier) = (method-qualifiers method)
        for method-function = (method-function method)
        if (eq qualifier :around) collect method-function into around-methods
        else if (eq qualifier :before) collect method-function into before-methods
        else if (eq qualifier :after) collect method-function into after-methods
        else collect method-function into primary-methods
        finally
        (unless primary-methods
          (method-combination-error
           "No applicable methods for generic function ~S with arguments ~S."
           function args))
        (let* ((inner-methods (if (or before-methods after-methods)
                                  (list (combine-methods before-methods 
                                                         primary-methods
                                                         after-methods))
                                primary-methods))
               (all-methods (nconc around-methods inner-methods)))
          (with-next-methods (rest all-methods)
                             (return (apply (first all-methods)
                                            args))))))

; Combine applicable methods
(defun combine-methods (before-methods primary-methods after-methods)
  (lambda (&rest args)
    (declare (dynamic-extent args))
    (with-next-methods '()
                       (loop for method in before-methods
                             do (apply method args))
                       (multiple-value-prog1
                           (with-next-methods (rest primary-methods)
                                              (apply (first primary-methods)
                                                     args))
                         (loop for method in after-methods
                               do (apply method args))))))

; Check whether two set of specialisers are the same.
(defun specificity->= (specialisers-1 specialisers-2)
  (loop for specialiser-1 in specialisers-1
        for specialiser-2 in specialisers-2
        if (and (not (listp specialiser-1))
                (not (listp specialiser-2)))
        unless (or (eql specialiser-1 specialiser-2)
                   (find-superclass specialiser-2
                                    (class-direct-superclasses specialiser-1)))
        return nil
        else
        if (and (listp specialiser-1)
                (listp specialiser-2))
        unless (eql (second specialiser-1)
                    (second specialiser-2))
        return nil
        else
        return nil
        finally return t))

; TO DO: To factorise out with the previous method
(defun specificity-= (specialisers-1 specialisers-2)
  (loop for specialiser-1 in specialisers-1
        for specialiser-2 in specialisers-2
        if (and (not (listp specialiser-1))
                (not (listp specialiser-2)))
        unless (eql specialiser-1 specialiser-2)
        return nil
        else
        if (and (listp specialiser-1)
                (listp specialiser-2))
        unless (eql (second specialiser-1)
                    (second specialiser-2))
        return nil
        else
        return nil
        finally return t))

; Find method of function with the specialisers passed as arguments
(defun find-method-with-specialisers (function specialisers)
  (find-if #'(lambda (method)
               (specificity-= specialisers
                              (method-specializers method)))
           (generic-function-methods function)))
               