;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  (__)  LAMBIC Project
;   )(   Written by Jorge Vallejos
;        
; (c) Vrije Universiteit Brussel, 2008 - 2010
; 
; Permission is hereby granted, free of charge, to any person
; obtaining a copy of this software and associated documentation
; files (the "Software"), to deal in the Software without
; restriction, including without limitation the rights to use,
; copy, modify, merge, publish, distribute, sublicense, and/or
; sell copies of the Software, and to permit persons to whom the
; Software is furnished to do so, subject to the following
; conditions:
;
; The above copyright notice and this permission notice shall be
; included in all copies or substantial portions of the Software.
;
; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
; EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
; OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
; NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
; HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
; WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
; OTHER DEALINGS IN THE SOFTWARE.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; LAMBIC CHANNELS
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package :lambic)

;;;;;;;;;;;;;;;;;;;;
; GENERIC CHANNELS ;
;;;;;;;;;;;;;;;;;;;;
(define-distributable-class generic-channel ()
  ((message-queue :initform nil :accessor channel-message-queue)
   (disconnection-observers :initform nil :accessor channel-disconnection-observers)
   (reconnection-observers :initform nil :accessor channel-reconnection-observers)
   (description :initform nil :initarg :description :accessor channel-description)))         

; Setting up timeout for a message
(defmethod set-timeout-for-message (channel message)
  (let ((timeout-name (format nil "Lambic-timer-~d-for-message:~d"
                              (incf *timer-id*)
                              (function-call-name (actor-message-function-call message)))))
    (mp:process-run-function
     timeout-name nil
     (lambda (channel message)
       (let ((timer (mp:make-timer 'send-to-actor-of channel
                                   'notify-message-timeout (list channel message)
                                   :with-future nil)))
         (mp:schedule-timer-relative timer (actor-message-timeout message))))
     channel message)))

; Notify message timeout. This method is called by
; the timer scheduled for the message.
(defmethod notify-message-timeout (channel message)
  (format *screen-output* "TIMEOUT CONDITION. call: ~d arguments: ~d~%"
          (function-call-name (actor-message-function-call message))
          (function-call-arguments (actor-message-function-call message)))
  (setf (channel-message-queue channel)
        (remove message (channel-message-queue channel)))
  (let ((future (actor-message-future message))
        (discard-value (actor-message-discard-value message)))
    (when future
      (if discard-value
          (in-actor-of future (resolve-with-discard-result future discard-value)
                       :with-future nil)
        (in-actor-of future (raise-timeout-condition future)
                     :with-future nil)))))

; Queue message
(defmethod queue-message (channel message)
  (push message (channel-message-queue channel)))

; Deliver actor message. This around method defines several 
; alternatives to handle the message in case that the channel
; is disconnected.
(defmethod deliver-actor-message :around (channel message)
  (if (channel-connectedp channel)
      ; if channel is connected, deliver message!
      (call-next-method)
    (let ((timeout (actor-message-timeout message))
          (discard-value (actor-message-discard-value message))
          (future (actor-message-future message)))
      (if timeout
          ; if message has timeout, set it up!
          (progn
            (queue-message channel message)
            (set-timeout-for-message channel message))
        (if discard-value
            ; if message has discard value, resolve the future
            ; with that value!
            (resolve-with-discard-result future discard-value)
          ; if message does not have discard value,
          ; queue message in channel
          (queue-message channel message))))))

; Deliver actor message. This method also defines alternatives
; to handle the message in case of timeout exceptions during
; the streaming.
(defmethod deliver-actor-message (channel message)
  (handler-case
      (stream-actor-message channel message)
    (error (e) (let ((discard-value (actor-message-discard-value message))
                    (future (actor-message-future message)))
                (disconnect channel)
                (if (actor-message-timeout message)
                    ; if message has timeout, notify timeout!
                    (progn
                      (error e)
                      (notify-message-timeout channel message))
                  (if discard-value
                      ; if message has discard value, resolve the future
                      ; with that value!
                      (resolve-with-discard-result future discard-value)
                    ; if message does not have discard value,
                    ; queue message in channel
                    (queue-message channel message)))))))

; Deliver queued messages
(defmethod deliver-queued-messages (channel)
  (let ((messages (reverse (copy-list (channel-message-queue channel)))))
    (dolist (message messages)
      (handler-case
          (progn
            (stream-actor-message channel message)
            (setf (channel-message-queue channel)
                  (remove message (channel-message-queue channel))))
        (error (e)
          (let ((discard-value (actor-message-discard-value message))
                (future (actor-message-future message)))
            (disconnect channel)
            (if (actor-message-timeout message)
                ; if message has timeout, notify timeout!
                (progn
                  (error e)
                  (notify-message-timeout channel message))
              (when discard-value
                ; if message has discard value, resolve the future
                ; with that value, and remove the message from the
                ; channel's queue!
                (resolve-with-discard-result future discard-value)
                (setf (channel-message-queue channel)
                      (remove message (channel-message-queue channel)))))
            (return nil)))))))

; Notify connection observers (of channel or of local host).
(defun notify-observers (object observers)
  (loop for observer in (reverse observers)
        do (in-actor-of observer (resolve-with-result observer object)
                        :with-future nil)))

; Add disconnection observer for a remote service (represented by channel).
(defmethod add-disconnection-observer (channel observer)
  (push observer (channel-disconnection-observers channel)))

; Add reconnection observer for a remote service (represented by channel).
(defmethod add-reconnection-observer (channel observer)
  (push observer (channel-reconnection-observers channel)))

; Print channel to a stream.
(defmethod print-object ((channel generic-channel) stream)
  (if *print-readably*
      (print (channel-reference channel) stream)
    (call-next-method)))

; Distinguish messages sent to the channel from those sent
; to the 
; This could be clearly improved
(defmethod meta-messagep (channel method-name)
  (let ((method-list (list 'CHANNEL-MESSAGE-QUEUE 'CHANNEL-DISCONNECTION-OBSERVERS
                           'CHANNEL-RECONNECTION-OBSERVERS 'CHANNEL-REFERENCE
                           'CHANNEL-CONNECTEDP 'CHANNEL-REFERENCES
                           'SET-TIMEOUT-FOR-MESSAGE 'NOTIFY-MESSAGE-TIMEOUT
                           'QUEUE-MESSAGE 'DELIVER-ACTOR-MESSAGE
                           'DELIVER-QUEUED-MESSAGES 'STREAM-ACTOR-MESSAGE
                           'RECONNECT 'DISCONNECT 'NOTIFY-OBSERVERS
                           'ADD-DISCONNECTION-OBSERVER 'ADD-RECONNECTION-OBSERVER
                           'PRINT-OBJECT)))
    (find method-name method-list :test #'equal)))

;;;;;;;;;;;;
; CHANNELS ;
;;;;;;;;;;;;
; By default a new channel is always in a connected state.
(define-distributable-class channel (generic-channel)
  ((far-reference :initarg :far-reference :accessor channel-reference)
   (connectedp :initform t :accessor channel-connectedp)))

; Stream actor message. Raise an error in case of
; connection problems.
(defmethod stream-actor-message ((channel channel) message)
  (with-standard-io-syntax
    (let ((*package* (find-package 'closer-common-lisp-user))
          (*print-circle* t))
     (with-open-stream
         (stream (comm:open-tcp-stream
                  (reference-address (channel-reference channel)) 
                  (reference-port (channel-reference channel))
                  :timeout (actor-message-timeout message)
                  :errorp t))
       (write message :stream stream :escape t :readably t)
       (force-output stream)))))

; Reconnect channel. Method invoked by the connection handler.
(defmethod reconnect ((channel channel) &optional address)
  (when (and (local-host-connectedp *communication-actor*)
             (not (channel-connectedp channel)))
    (setf (channel-connectedp channel) t)
    (deliver-queued-messages channel)
    (notify-observers channel (channel-reconnection-observers channel))))

; Disconnect channel. Method invoked by the connection handler.
(defmethod disconnect ((channel channel) &optional address)
  (when (channel-connectedp channel)
    (setf (channel-connectedp channel) nil)
    (when (local-host-connectedp *communication-actor*)
      (notify-observers channel (channel-disconnection-observers channel)))))

;;;;;;;;;;;;;;;;;;
; GROUP CHANNELS ;
;;;;;;;;;;;;;;;;;;

(define-distributable-class group-channel (generic-channel)
  ((far-references :initform nil :initarg :far-references :accessor channel-references)))

(defmethod channel-connectedp ((channel group-channel))
  (find-if #'(lambda (reference)
               (car reference))
           (channel-references channel)))

; Reconnect channel. Method invoked by the connection handler.
(defmethod reconnect ((channel group-channel) &optional address)
  (when (and (local-host-connectedp *communication-actor*)
             (not (channel-connectedp channel)))
    (let ((reference (find-if #'(lambda (reference)
                                  (equal (reference-address (cdr reference))
                                         address))
                              (channel-references channel))))
      (setf (car reference) t)
      (deliver-queued-messages channel)
      (notify-observers channel (channel-reconnection-observers channel)))))


; Disconnect channel. Method invoked by the connection handler.
(defmethod disconnect ((channel group-channel) &optional address)
  (when (channel-connectedp channel)
    (let ((dead-reference (find-if #'(lambda (reference)
                                       (equal (reference-address (cdr reference))
                                         address))
                                   (channel-references channel))))
      (setf (car dead-reference) nil)
      (unless (and (channel-connectedp channel)
                   (not (local-host-connectedp *communication-actor*)))
          (notify-observers channel (channel-disconnection-observers channel))))))


; Stream actor message. Raise an error in case of
; connection problems.
(defmethod stream-actor-message ((channel group-channel) message)
  (with-standard-io-syntax
    (let ((*package* (find-package 'closer-common-lisp-user))
          (*print-circle* t))
      (loop for reference in (channel-references channel)
            for i from 1 to (length (channel-references channel))
            unless (handler-case
                       (with-open-stream
                           (stream (comm:open-tcp-stream
                                    (reference-address reference) 
                                    (reference-port reference)
                                    :timeout (actor-message-timeout message)
                                    :errorp t))
                         (write message :stream stream :escape t :readably t)
                         (force-output stream))
                     (error (var) 
                       (when (eql i (length (channel-references channel)))
                         (error var))))
            return t))))

;;;;;;;;;;;;;;;;;;;
; ACTOR EXTENSION ;
;;;;;;;;;;;;;;;;;;;
; send-actor-message method definition specialised on channels
(defmethod send-actor-message ((this actor) (receiver generic-channel) function-call future
                               &key due-in discard-with-value)
  (let ((message (make-actor-message :sender this :receiver receiver
                                     :function-call function-call :future future
                                     :timeout due-in :discard-value discard-with-value)))
    (if (meta-messagep receiver (function-call-name function-call))
        (receive-actor-message (object-actor receiver) message)
      (in-actor-of receiver (deliver-actor-message receiver message)
                   :with-future nil))
    future))


(defmethod actor-of ((object generic-channel))
  (reference-actor-name (channel-reference object)))

; Return NIL for far references.
(defmethod in-current-actor ((object generic-channel))
  nil)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; REMOTE CONNECTION CALLBACKS ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; DUMMY! To improve design here.
; Add disconnection observer for a local service (represented by
; distributable-object).
; This method cannot just be removed as this callback is commonly
; used inside a whenever-discovered callback, which can be resolved
; with local objects. See the chat application as example. 
(defmethod when-disconnected ((service distributable-object) continuation-lambda)
  "Disconnection will presumably never happen")

; Install an observer for disconnection of a channel
(defmethod when-disconnected ((service generic-channel) continuation-lambda)
  (let ((disconnection-future (make-instance 'future)))
    (when-resolved disconnection-future continuation-lambda)
    (in-actor-of service
                 (add-disconnection-observer service disconnection-future)
                 :with-future nil)))

; DUMMY! To improve design here.
; Add reconnection observer for a local service (represented by
; distributable-object).
; This method cannot just be removed as this callback is commonly
; used inside a whenever-discovered callback, which can be resolved
; with local objects. See the chat application as example.
(defmethod when-reconnected ((service distributable-object) continuation-lambda)
  "Disconnection (and thus reconnection) will presumably never happen")

; Install an observer for reconnection of a channel 
(defmethod when-reconnected ((service generic-channel) continuation-lambda)
  (let ((reconnection-future (make-instance 'future)))
    (when-resolved reconnection-future continuation-lambda)
    (in-actor-of service
                 (add-reconnection-observer service reconnection-future)
                 :with-future nil)))



