;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  (__)  LAMBIC Project
;   )(   Written by Jorge Vallejos
;        
; (c) Vrije Universiteit Brussel, 2008 - 2010
; 
; Permission is hereby granted, free of charge, to any person
; obtaining a copy of this software and associated documentation
; files (the "Software"), to deal in the Software without
; restriction, including without limitation the rights to use,
; copy, modify, merge, publish, distribute, sublicense, and/or
; sell copies of the Software, and to permit persons to whom the
; Software is furnished to do so, subject to the following
; conditions:
;
; The above copyright notice and this permission notice shall be
; included in all copies or substantial portions of the Software.
;
; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
; EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
; OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
; NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
; HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
; WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
; OTHER DEALINGS IN THE SOFTWARE.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; LAMBIC COMMUNICATION
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package :lambic)

;;;;;;;;;;;;;;;;;;;;;;;;;;;
; COMMUNICATION BEHAVIOUR ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;
(define-distributable-class communicator ()
  ((channels :initform nil :accessor communicator-channels)
   (local-host-connectedp :initform t :accessor local-host-connectedp)
   (local-disconnection-observers :initform nil :accessor local-disconnection-observers)
   (local-reconnection-observers :initform nil :accessor local-reconnection-observers)))

; Add observer for local host's disconnections.
(defmethod add-local-disconnection-observer (communicator service observer)
  (push (cons service observer) (local-disconnection-observers communicator)))

; Add observer for local host's reconnections.
(defmethod add-local-reconnection-observer (communicator service observer)
  (push (cons service observer) (local-reconnection-observers communicator)))

; Notify disconnection of host.
(defmethod notify-local-disconnection (communicator)
  (when (local-host-connectedp communicator)
    (setf (local-host-connectedp communicator) nil)
    (notify-local-connection-observers (local-disconnection-observers communicator))))

; Notify reconnection of host.
(defmethod notify-local-reconnection (communicator)
  (unless (local-host-connectedp communicator)
    (setf (local-host-connectedp communicator) t)
    (notify-local-connection-observers (local-reconnection-observers communicator))))

; Notify observers of local connection.
(defun notify-local-connection-observers (observers)
  (loop for observer-cons in (reverse observers)
        do (in-actor-of (cdr observer-cons)
                        (resolve-with-result (cdr observer-cons)
                                             (car observer-cons))
                        :with-future nil)))

; Unique way to send message using channels
(defmethod send-remote-actor-message (communicator message)
  (let* ((receiver-reference (actor-message-receiver message))
         (channel (provide-channel communicator receiver-reference)))
    (in-actor-of channel
                 (deliver-actor-message channel message)
                 :with-future nil)))

(defmethod provide-channel (communicator far-reference)
  (let ((channel (find-channel communicator far-reference)))
    (unless channel
      ; Create the channel to the remote actor if it does not exist yet.
      (setf channel (create-channel communicator far-reference)))
    channel))

; Find a channel
(defmethod find-channel (communicator far-reference)
  (dolist (channel (communicator-channels communicator))
    (and (if (typep channel 'group-channel)
             (find-if #'(lambda (reference-cons)
                          (reference-equal (cdr reference-cons) far-reference))
                      (channel-references channel))
           (reference-equal (channel-reference channel) far-reference))
         (return channel))))

; Create a channel and add it to the communication actor
(defmethod create-channel (communicator (reference far-reference)
                                        &optional (service-description nil))
  (let ((channel (make-instance 'channel
                                :far-reference reference)))
    (spawn-actor (generate-new-actor-name) channel)
    (push channel (communicator-channels communicator))
    channel))

; Get or create a group channel
(defmethod create-channel (communicator (reference group-far-reference)
                                        &optional (service-description nil))
  (let ((group-channel (or (find-if #'(lambda (channel)
                                        (when (channel-description channel)
                                          (description-record-equal
                                           (channel-description channel)
                                           service-description)))
                                    (communicator-channels communicator))
                           (make-instance 'group-channel
                                          :description service-description)))
        (channel (call-next-method)))
    (unless (channel-references group-channel)
      (spawn-actor (generate-new-actor-name) group-channel)
      (push group-channel (communicator-channels communicator)))
    (push (cons t reference) (channel-references group-channel))
    (cons group-channel channel)))

;;;;;;;;;;;;;;;;;;;;;;;;
; COMMUNICATION SERVER ;
;;;;;;;;;;;;;;;;;;;;;;;;

; Initialisation of communication server
(defmethod start-communication (communicator)
  (comm:start-up-server-and-mp 
   :function (lambda (port)
               (let ((stream (make-instance 'comm:socket-stream :socket port
                                            :direction :input :element-type 'base-char)))
                 (mp:process-run-function (format nil "Lambic-comm-server-~D" port)
                  '()
                  'read-message-from-stream stream)))
   :service *device-port*
   :keepalive t
   :process-name (format nil "Lambic-communication-server-at-port-~D" *device-port*))
  (start-connection-handler communicator))

(defvar *ping-message* "ping")

; Function to convert message received from stream
; into an actor message
(defun read-message-from-stream (stream)
  (unwind-protect
      (with-standard-io-syntax
        (let* ((*package* (find-package 'closer-common-lisp-user))
               (actor-message (read stream)))
          (unless (equal actor-message *ping-message*)
            (receive-actor-message (object-actor (actor-message-receiver actor-message))
                                   actor-message))))
    (close stream)))

;;;;;;;;;;;;;;;;;;;;;;
; CONNECTION HANDLER ;
;;;;;;;;;;;;;;;;;;;;;;

; This process (not an actor) notifies changes of connection status
; to the channels.
(defun start-connection-handler (communicator)
  (mp:process-run-function
   "Lambic-connection-handler-process" nil
   (lambda ()
     (loop
      (let ((local-host-connectedp (check-local-connection)))
        (if local-host-connectedp
            (let* ((channels-copy (communicator-channels communicator))
                   (current-channels (make-host-dictionary channels-copy)))
              (in-actor-of communicator (notify-local-reconnection communicator)
                           :with-future nil)
              (maphash #'(lambda (host-address host-channels)
                           (loop for channel in host-channels
                                 if (check-remote-connection host-address *device-port*)
                                 do (in-actor-of channel
                                                 (reconnect channel host-address)
                                                 :with-future nil)
                                 else
                                 do (in-actor-of channel
                                                 (disconnect channel host-address)
                                                 :with-future nil)))
                       current-channels))
          (in-actor-of communicator (notify-local-disconnection communicator)
                       :with-future nil)))
      (mp:process-wait-with-timeout "Waiting" 5)))))

; Make hosts dictionary to avoid multiple pings to the same host
; in one pinging round.
(defun make-host-dictionary (channels)
  (let ((dictionary (make-hash-table :test 'equal)))
    (dolist (channel channels)
      (update-dictionary channel dictionary))
    dictionary))

; Update the dictionary with the channel
(defmethod update-dictionary ((channel channel) dictionary)
  (let ((address (reference-address (channel-reference channel))))
    (push channel (gethash address dictionary))))    

; Update the dictionary with the group channel. Register the channel
; for each of the references' adrress.
(defmethod update-dictionary ((channel group-channel) dictionary)
  (dolist (reference (channel-references channel))
    (push channel (gethash (reference-address (cdr reference))
                           dictionary))))

; Check if local connection is alive
(defun check-local-connection ()
  (handler-case
      (with-open-stream (http (comm:open-tcp-stream "www.lispworks.com" 80
                                                    :errorp t :timeout 3))
        (force-output http)
        http)
    (error () nil)))

; Check if connection to a remote channel is alive
(defun check-remote-connection (host-address host-port)
  (handler-case
      (with-standard-io-syntax
        (let ((*package* (find-package 'closer-common-lisp-user))
              (*print-circle* t))
          (with-open-stream
              (stream (comm:open-tcp-stream host-address host-port
                                            :timeout 3 :errorp t))
            (write *ping-message* :stream stream :escape t :readably t)
            (force-output stream)
            stream)))
    (error () nil)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; LOCAL CONNECTION CALLBACKS ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; Install observer for disconnection of local host.
(defmethod when-self-disconnected ((service distributable-object) continuation-lambda)
  (let ((disconnection-future (make-instance 'future)))
    (when-resolved disconnection-future continuation-lambda)
    (in-actor-of *communication-actor*
                 (add-local-disconnection-observer *communication-actor*
                                                   service disconnection-future)
                 :with-future nil)))

; Install observer for reconnection of local host.
(defmethod when-self-reconnected ((service distributable-object) continuation-lambda)
  (let ((reconnection-future (make-instance 'future)))
    (when-resolved reconnection-future continuation-lambda)
    (in-actor-of *communication-actor*
                 (add-local-reconnection-observer *communication-actor*
                                                   service reconnection-future)
                 :with-future nil)))


; INITIALISE *communication-actor*
(setf *communication-actor* (make-instance 'communicator))
(spawn-actor "communicator" *communication-actor*)
(start-communication *communication-actor*)

