;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  (__)  LAMBIC Project
;   )(   Written by Jorge Vallejos
;        
; (c) Vrije Universiteit Brussel, 2008 - 2010
; 
; Permission is hereby granted, free of charge, to any person
; obtaining a copy of this software and associated documentation
; files (the "Software"), to deal in the Software without
; restriction, including without limitation the rights to use,
; copy, modify, merge, publish, distribute, sublicense, and/or
; sell copies of the Software, and to permit persons to whom the
; Software is furnished to do so, subject to the following
; conditions:
;
; The above copyright notice and this permission notice shall be
; included in all copies or substantial portions of the Software.
;
; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
; EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
; OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
; NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
; HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
; WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
; OTHER DEALINGS IN THE SOFTWARE.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; LAMBIC CONTEXT PREDICATES
;
; MOP extension to enable:
;
; - generic functions to contain predicates
; - methods to use those predicates
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package :lambic)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; PREDICATED GENERIC FUNCTION DEFINITION ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar *max-predicates* 10)

; DEFINE-PREDICATED-GENERIC macro
(defmacro define-predicated-generic (function-name (&rest lambda-list)
                                                   &body options)
  (multiple-value-bind (options predicates)
      (remove-if* #'(lambda (option) (eql (car option) :predicates))
                  options :return-removed-items t)
    (let ((udpated-lambda-list (update-lambda-list lambda-list)))
      `(prog1
           (define-futurised-generic ,function-name ,udpated-lambda-list
             ,@(when options `(,options)))
           ;(predicate-function-method-class function)
         ,@(when predicates
             `((progn
                (setf (function-predicates (fdefinition ',function-name))
                      ',(parse-function-predicates (cdar predicates))))))))))
;               TODO: To be re-evaluated the need for the line below
;               (update-methods (c2cl:generic-function-methods function))))))))))

; Update the function's lambda list with predicates arguments
(defun update-lambda-list (lambda-list)
  (append lambda-list
          (loop for i from 1 to *max-predicates*
                collect (create-predicate-argument-name))))

; Create a gensym-based name for a predicate argument
(defun create-predicate-argument-name ()
  (read-from-string (format nil "predicate-~d" (gensym))))



; Convert predicates form into a hash table of predicate functions
; using the predicate's name as key.
; E.g. for a definition of predicates:
; ('foo (bar (a) ...) (baz (x y) ...))
; produce:
; '(foo (bar lambda) (baz lambda))
(defun parse-function-predicates (predicates-form)
  (loop for form in predicates-form
        if (eql (car form) 'quote)
        collect (eval form)
        else
        collect (list (car form)
                      (progn 
                        (eval `(defmethod ,@form))
                        (fdefinition (car form))))))

#|
; Let methods of a generic function be instances of the
; PREDICATED-METHOD class
(defun predicate-function-method-class (function)
  (let* ((method-class (c2cl:generic-function-method-class function))
         (standard-method-class (find-class 'standard-method))
         (predicated-class (find-class 'predicated-method)))
    (if (equal method-class standard-method-class)
        (setf (c2cl:generic-function-method-class function)
              predicated-class)
      (add-superclass predicated-class method-class))))
|#


; Update specialisers of the existing methods of generic functions.
; This is in case that the generic function is being redefined.
(defun update-methods (methods)
  (loop for method in methods
        do (update-method-specialisers method)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; PREDICATED METHOD DEFINITION ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;



#|
; DEFINE-PREDICATED-METHOD macro
(defmacro define-predicated-method (&rest forms)
  (multiple-value-bind (name qualifiers lambda-list declarations
                             documentation body argument-names specialisers
                             method-predicates)
      (parse-predicated-method-definition forms)
    (declare (ignore specialisers argument-names))
    (when (and method-predicates (not (fboundp name)))
      (error "Predicates should be specified in the generic function first."))
    (let ((updated-lambda-list (update-lambda-list lambda-list)))
      `(progn
         #|(unless (fboundp ',name)
           (define-predicated-generic ,name ,argument-names))|#
         (let ((method (define-futurised-method ,name
                         ,@(when qualifiers `(,@qualifiers))
                         ,updated-lambda-list
                         ,@(when declarations `(,@declarations))
                         ,@(when documentation `(,documentation))
                         ,@body)))
           (setf (method-predicates method) ',(cdar method-predicates))
           (update-method-specialisers method)
           method)))))
|#

; DEFINE-PREDICATED-METHOD macro
(defmacro define-predicated-method (&rest forms)
  (multiple-value-bind (name qualifiers lambda-list declarations
                             documentation body argument-names specialisers
                             method-predicates)
      (parse-predicated-method-definition forms)
    (declare (ignore argument-names))
    (let* ((updated-specialisers (append specialisers
                                         (get-context-specialisers name
                                                                   (cdar method-predicates))))
           (updated-lambda-list (update-specialised-lambda-list lambda-list
                                                                updated-specialisers)))
      `(prog1
         (define-futurised-method ,name
                                  ,@(when qualifiers `(,@qualifiers))
                                  ,updated-lambda-list
                                  ,@(when declarations `(,@declarations))
                                  ,@(when documentation `(,documentation))
                                  ,@body)
         ,@(when method-predicates
             `((let ((method (find-method-with-specialisers (fdefinition ',name)
                                                           ',updated-specialisers)))
                (setf (method-predicates method) ',(cdar method-predicates)))))))))

; Parse a predicated method definition
(defun parse-predicated-method-definition (forms)
  (multiple-value-bind (name qualifiers specialised-lambda-list declarations
                             documentation body argument-names original-specialisers)
      (parse-method-definition forms)
    (multiple-value-bind (body predicates)
        (remove-if* #'(lambda (expression)
                        (when (listp expression) (eql (car expression) :when)))
                    body :return-removed-items t)
      (values name qualifiers specialised-lambda-list declarations
              documentation body argument-names original-specialisers predicates))))

; Update lambda list with context predicate classes.
(defun update-specialised-lambda-list (lambda-list new-specialisers)
  (let ((new-specialiser-names (get-specialiser-names new-specialisers)))
    (append lambda-list
            (loop for i 
                  from (length lambda-list)
                  to (1- (+ (length lambda-list) *max-predicates*))
                  collect (list (create-predicate-argument-name)
                                (nth i new-specialiser-names))))))


#|
(defun update-lambda-list (lambda-list)
  (append lambda-list
          (loop for i from 1 to *max-predicates*
                collect (create-predicate-argument-name))))
|#

; Determine the specialisers corresponding to the predicates
; used in a method. Update the specialisers list and store the
; predicates in the method.
(defun update-method-specialisers (method)
  (let* ((method-predicates (method-predicates method))
         (function (method-generic-function method))
         specialisers)
    (dotimes (i 10)
      (let ((predicate (nth i method-predicates)))
        (push (if predicate 
                  (get-predicate-class function (car predicate))
                (find-class 'lambic-predicate))
              specialisers)))
    (setf specialisers (reorder-predicate-specialisers specialisers))
    (setf (method-specializers method) 
          (replace-predicate-specialisers (method-specializers method)
                                          specialisers))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;






;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


; Get the Lambic predicate class that corresponds to a name
(defmethod get-predicate-class ((function predicated-function) predicate-name)
  (let ((position (position predicate-name (function-predicates function)
                            :key #'(lambda (predicate)
                                     (if (typep predicate 'list)
                                         (car predicate)
                                       predicate)))))
    (unless position
      (error "Predicate ~d is not specified in generic function ~d."
               predicate-name function))
    (nth position *lambic-predicate-classes*)))

; Reorder predicate specialisers using subclass relationship
; of lambic predicate classes.
(defun reorder-predicate-specialisers (specialisers)
  (let (ordered-specialisers)
    (loop for specialiser in specialisers
          do (setf ordered-specialisers
                   (add-specialiser-in-order specialiser
                                             ordered-specialisers)))
    ordered-specialisers))

; Add specialiser in order (auxiliar function of previous
; function)
(defun add-specialiser-in-order (specialiser specialisers)
  (if specialisers
      (if (subtypep specialiser (car specialisers))
          (append (list specialiser) specialisers)
        (append (list (car specialisers))
                (add-specialiser-in-order specialiser (cdr specialisers))))
    (list specialiser)))

; Append the predicate specialisers to the argument specialisers
(defun replace-predicate-specialisers (method-specialisers predicate-specialisers)
  (let* ((arguments-number (- (length method-specialisers)
                              (length predicate-specialisers)))
         (argument-specialisers (loop for i from 0 to (1- arguments-number)
                                      collect (nth i method-specialisers))))
    (append argument-specialisers predicate-specialisers)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; PREDICATED FUNCTION EXECUTION ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod compute-discriminating-function ((function predicated-function))
  (lambda (&rest args)
    (let* ((result-future (make-instance 'future))
           (predicate-args (make-list (- (length (generic-function-lambda-list function))
                                         (length args))
                                      :initial-element (make-instance 'lambic-predicate-9)))
           (extended-args (append args predicate-args))
           (original-methods (compute-applicable-methods function extended-args))
           (future-list (futurised-remove-if #'(lambda (method)
                                                 (eval-method-predicates method args))
                                             original-methods)))
      (when-resolved future-list
        (lambda (selected-methods)
          (resolve-with-result result-future
                               (apply-selected-methods selected-methods function
                                                       extended-args))))
      result-future)))

; Eval argument in method's extended environment. Eval
; argument in standard environment otherwise. 
(defmacro context-eval (arg env)
  `(or (prog1 (gethash ,arg ,env)) (eval ,arg)))

; Eval method predicates
(defmethod eval-method-predicates (method parameters)
  (let ((method-predicates (and (typep method 'predicated-method)
                                (method-predicates method))))
    (when method-predicates
      (let ((env (make-hash-table))
            (function (method-generic-function method)))
        ;(debug-print "method-predicates" method-predicates)
        (add-predicates-to-environment function env)
        (add-parameters-to-environment method parameters env)
        ;(debug-print "env" env)
        (dolist (predicate method-predicates)
          (let ((predicate-function (context-eval (car predicate) env))
                (predicate-parameters (loop for name in (cdr predicate)
                                            collect (context-eval name env))))
            ;(debug-print "predicate-function" predicate-function)
            (futurised-unless (apply predicate-function predicate-parameters)
              (return t))))))))

; Add generic function's predicates to the method's execution
; envionrment
(defun add-predicates-to-environment (function env)
  (let ((function-predicates (function-predicates function)))
    (dolist (predicate function-predicates)
      (if (typep predicate 'list)
          (setf (gethash (first predicate) env) (second predicate))
        (let ((external-predicate (fdefinition predicate)))
          (unless external-predicate
            (error "Unexisting function ~d used as predicate in generic function ~d"
                   predicate function))
          (setf (gethash predicate env) external-predicate))))))

; Add method's parameters to the method's execution environment
(defun add-parameters-to-environment (method parameters env)
  (loop for name in (method-lambda-list method)
        for value in parameters
        do (setf (gethash name env) value)))

;;;;;;;;;
; TESTS ;
;;;;;;;;;

#|

(in-package :lambic)

(defvar f1 (make-instance 'future))
(defvar f2 (make-instance 'future))
(define-futurised-method my-oddp (v)
                         (oddp v))
(when-resolved (futurised-remove-if #'(lambda (v)
                                        (my-oddp v))
                                    (list f1 f2))
               (lambda (result)
                 (format *screen-output* "Resulting list:~d" result)))

(resolve-with-result f1 1)
(resolve-with-result f2 2)


(in-package :lambic)

(define-predicated-generic foo (a b c)
                            (:predicates
                             (bar (a) (> 0))
                             'my-oddp))

;(define-predicated-generic my-oddp (v))

(define-predicated-method my-oddp (v)
                         (oddp v))

(define-predicated-method foo (a b c)
  (print "bottom"))

(define-predicated-method foo (a b c)
                          (:when (bar a))
  (print "a is > 0")
  (call-next-method))

(define-predicated-method foo (a b c)
                          (:when (my-oddp b))
  (print "a is > 0 and b is an odd number")
  (call-next-method))

(foo 1 2 3)

(foo 1 3 2)
|#   
