;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  (__)  LAMBIC Project
;   )(   Written by Jorge Vallejos
;        
; (c) Vrije Universiteit Brussel, 2008 - 2010
; 
; Permission is hereby granted, free of charge, to any person
; obtaining a copy of this software and associated documentation
; files (the "Software"), to deal in the Software without
; restriction, including without limitation the rights to use,
; copy, modify, merge, publish, distribute, sublicense, and/or
; sell copies of the Software, and to permit persons to whom the
; Software is furnished to do so, subject to the following
; conditions:
;
; The above copyright notice and this permission notice shall be
; included in all copies or substantial portions of the Software.
;
; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
; EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
; OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
; NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
; HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
; WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
; OTHER DEALINGS IN THE SOFTWARE.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; LAMBIC CONTEXT PREDICATES
;
; MOP extension to enable:
;
; - generic functions to contain predicates
; - methods to use those predicates
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package :lambic)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; AUXILIARS FOR DEFINE-PEER-GENERIC ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; Extract predicates option. Return the predicates and the other
; options separately.
(defun extract-predicates-option (options)
  (multiple-value-bind (other-options predicates)
      (remove-if* #'(lambda (option) (eql (car option) :predicates))
                  options :return-removed-items t)
    (values other-options (cdar predicates))))

; Convert predicates form into a list of predicate functions
; using the predicate's name as key.
; E.g. for a definition of predicates:
; ('foo (bar (a) ...) (baz (x y) ...))
; produce:
; '(foo (bar lambda) (baz lambda))
(defun parse-function-predicates (predicates-form)
  (loop for form in predicates-form
        if (eql (car form) 'quote)
        collect (eval form)
        else
        collect `(,(car form)
                  ,(eval `(lambda ,@(cdr form))))))

#|
(defun parse-function-predicates (predicates-form)
  (loop for form in predicates-form
        if (eql (car form) 'quote)
        collect (eval form)
        else
        collect (list (car form)
                      (progn 
                        (eval `(lambda ,@(cdr form)))
                        (fdefinition (car form))))))
|#


; Add predicates arguments to the generic function lambda list. 
(defun update-peer-function-lambda-list (lambda-list)
  (append lambda-list
          (loop for i from 1 to *max-predicates*
                collect (create-predicate-argument-name))))

; Create a gensym-based name for a predicate argument
(defun create-predicate-argument-name ()
  (read-from-string (format nil "predicate-~d" (gensym))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; AUXILIARS FOR DEFINE-PEER-METHOD ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; Extract context predicates from body
(defun extract-context-predicates (body)
  (multiple-value-bind (body predicates)
      (remove-if* #'(lambda (expression)
                      (when (listp expression) (eql (car expression) :when)))
                  body :return-removed-items t)
    (values body (cdar predicates))))

; Update lambda list with context predicate classes.
(defun update-specialised-lambda-list (lambda-list new-specialisers)
  (let ((new-specialiser-names (get-specialiser-names new-specialisers)))
    (append lambda-list
            (loop for i 
                  from (length lambda-list)
                  to (1- (+ (length lambda-list) *max-predicates*))
                  collect (list (create-predicate-argument-name)
                                (nth i new-specialiser-names))))))

; Get the name of the method specialisers
(defun get-specialiser-names (specialisers)
  (loop for specialiser in specialisers
        if (listp specialiser)
        collect specialiser
        else
        collect (class-name specialiser)))

; Get context specialisers
(defun get-context-specialisers (method-name method-predicates)
  (when (and method-predicates (not (fboundp method-name)))
    (error "Predicates should be specified in the generic function first."))
  (let ((specialisers (loop for i from 0 to 9
                            if (nth i method-predicates)
                            collect (get-predicate-class (fdefinition method-name)
                                                         (car (nth i method-predicates)))
                            else
                            collect (find-class 'lambic-predicate))))
    (reorder-predicate-specialisers specialisers)))

; Reorder predicate specialisers using subclass relationship
; of lambic predicate classes.
(defun reorder-predicate-specialisers (specialisers)
  (let (ordered-specialisers)
    (loop for specialiser in specialisers
          do (setf ordered-specialisers
                   (add-specialiser-in-order specialiser
                                             ordered-specialisers)))
    ordered-specialisers))

; Add specialiser in order (auxiliar function of previous
; function)
(defun add-specialiser-in-order (specialiser specialisers)
  (if specialisers
      (if (subtypep specialiser (car specialisers))
          (append (list specialiser) specialisers)
        (append (list (car specialisers))
                (add-specialiser-in-order specialiser (cdr specialisers))))
    (list specialiser)))

; Get the Lambic predicate class that corresponds to a name
(defmethod get-predicate-class ((function predicated-function) predicate-name)
  (let ((position (position predicate-name (function-predicates function)
                            :key #'(lambda (predicate)
                                     (if (typep predicate 'list)
                                         (car predicate)
                                       predicate)))))
    (unless position
      (error "Predicate ~d is not specified in generic function ~d."
               predicate-name function))
    (nth position *lambic-predicate-classes*)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; PREDICATED FUNCTION EXECUTION ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod compute-discriminating-function ((function predicated-function))
  (lambda (&rest args)
    (let* ((result-future (make-instance 'future))
           (predicate-args (make-list (- (length (generic-function-lambda-list function))
                                         (length args))
                                      :initial-element (make-instance 'lambic-predicate-9)))
           (extended-args (append args predicate-args))
           (original-methods (compute-applicable-methods function extended-args))
           (future-list (futurised-remove-if #'(lambda (method)
                                                 (eval-method-predicates method args))
                                             original-methods)))
      (when-resolved future-list
                     (lambda (selected-methods)
                       (resolve-with-result result-future
                                            (apply-selected-methods selected-methods function
                                                                    extended-args))))
      result-future)))

(defmethod compute-applicable-predicated-methods (function original-args)
  (let* ((predicate-args (make-list (- (length (generic-function-lambda-list function))
                                         (length original-args))
                                      :initial-element (make-instance 'lambic-predicate-9)))
         (extended-args (append original-args predicate-args))
         (original-methods (compute-applicable-methods function extended-args)))
    (futurised-remove-if (lambda (method)
                           (eval-method-predicates method original-args))
                         original-methods)))

#|
(defmethod compute-discriminating-function ((function predicated-function))
  (lambda (&rest args)
    (let* ((result-future (make-instance 'future))
           (predicate-args (make-list (- (length (generic-function-lambda-list function))
                                         (length args))
                                      :initial-element (make-instance 'lambic-predicate-9)))
           (extended-args (append args predicate-args))
           (original-methods (compute-applicable-methods function extended-args))
           (future-list (futurised-remove-if #'(lambda (method)
                                                 (eval-method-predicates method args))
                                             original-methods)))
      (when-resolved future-list
        (lambda (selected-methods)
          (resolve-with-result result-future
                               (apply-selected-methods selected-methods function
                                                       extended-args))))
      result-future)))
|#

; Eval argument in method's extended environment. Eval
; argument in standard environment otherwise. 
(defmacro context-eval (arg env)
  `(multiple-value-bind (value presentp)
           (gethash ,arg ,env)
     (if presentp
         value
       (eval ,arg))))

; Eval method predicates
(defmethod eval-method-predicates (method parameters)
  (let ((method-predicates (and (typep method 'predicated-method)
                                (method-predicates method))))
    (when method-predicates
      (let ((env (make-hash-table))
            (function (method-generic-function method)))
        ;(debug-print "method-predicates" method-predicates)
        (add-predicates-to-environment function env)
        (add-parameters-to-environment method parameters env)
        ;(debug-print "env" env)
        (dolist (predicate method-predicates)
          (let ((predicate-function (context-eval (car predicate) env))
                (predicate-parameters (loop for name in (cdr predicate)
                                            collect (context-eval name env))))
            ;(debug-print "predicate-function" predicate-function)
            (futurised-unless (apply predicate-function predicate-parameters)
              (return t))))))))

; Add generic function's predicates to the method's execution
; envionrment
(defun add-predicates-to-environment (function env)
  (let ((function-predicates (function-predicates function)))
    (dolist (predicate function-predicates)
      (if (typep predicate 'list)
          (setf (gethash (first predicate) env) (second predicate))
        (let ((external-predicate (fdefinition predicate)))
          (unless external-predicate
            (error "Unexisting function ~d used as predicate in generic function ~d"
                   predicate function))
          (setf (gethash predicate env) external-predicate))))))

; Add method's parameters to the method's execution environment
(defun add-parameters-to-environment (method parameters env)
  (loop for name in (method-lambda-list method)
        for value in parameters
        do (setf (gethash name env) value)))

