;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  (__)  LAMBIC Project
;   )(   Written by Jorge Vallejos
;        
; (c) Vrije Universiteit Brussel, 2008 - 2010
; 
; Permission is hereby granted, free of charge, to any person
; obtaining a copy of this software and associated documentation
; files (the "Software"), to deal in the Software without
; restriction, including without limitation the rights to use,
; copy, modify, merge, publish, distribute, sublicense, and/or
; sell copies of the Software, and to permit persons to whom the
; Software is furnished to do so, subject to the following
; conditions:
;
; The above copyright notice and this permission notice shall be
; included in all copies or substantial portions of the Software.
;
; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
; EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
; OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
; NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
; HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
; WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
; OTHER DEALINGS IN THE SOFTWARE.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; LAMBIC DISCOVERY
;
; The Lambic's discovery mechanism is based on the library 
; for Zeroconf service discovery, CL-ZEROCONF [John Wiseman 2005].
; This library makes the Apple Bonjour framework available to
; Common Lisp programs.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package :lambic)

; cl-zeroconf browser
(defvar *cl-zeroconf-browser*)

; cl-zeroconf-browser is a distributable class
; so that it can be shield by an actor.
(define-distributable-class cl-zeroconf-browser () ())

;;;;;;;;;;;;;;;;;;;;
; DISCOVERER CLASS ;
;;;;;;;;;;;;;;;;;;;;
(define-distributable-class discoverer ()
  ((required-services :initform nil :accessor required-services)
   #|e.g. ((service-description-1 . :one) (service-description-2 . :any))|#
   (imported-services :initform nil :accessor imported-services)
   #|e.g. ((service-description-1 . far-reference-1)
           (service-description-2 . distributable-object-2))|#
   (published-services :initform nil :accessor published-services)
   #|e.g. ((service-description-1 . dns-sd-service-1))|#
   (whenever-observers :initform nil :accessor whenever-observers)
   #|((service-description-1 . observer-1) (service-description-2 . observer-2))|#
   (group-whenever-observers :initform nil :accessor group-whenever-observers)))

; Check whether a remote service reference was already discovered
(defmethod find-imported-service (discoverer (remote-service far-reference)
                                             service-description)
  (find-if #'(lambda (stored-cons)
                 (and (typep (cdr stored-cons) 'far-reference)
                      (reference-equal (cdr stored-cons) remote-service)
                      (description-record-equal service-description (car stored-cons))))
           (imported-services discoverer)))

; Check whether a local service reference was already discovered
(defmethod find-imported-service (discoverer (local-service distributable-object)
                                             service-description)
  (find-if #'(lambda (stored-cons)
                 (and (typep (cdr stored-cons) 'distributable-object)
                      (eql (cdr stored-cons) local-service)
                      (description-record-equal service-description (car stored-cons))))
           (imported-services discoverer)))

; Check whether a remote service reference is required
(defmethod find-required-service (discoverer discovered-description)
  (find-if #'(lambda (required-cons)
               (description-record-equal (car required-cons)
                                         discovered-description))
           (required-services discoverer)))

; Remove required service description after importing
; a service with such description. Keep the service description
; if it was added using the whenever-discovered construct. 
(defmethod remove-required-description-if-applicable (discoverer service-description)
  (setf (required-services discoverer)
        (remove-if #'(lambda (required-cons)
                       (and (description-record-equal service-description (car required-cons))
                            (not (equal (cdr required-cons) :any))))
                   (required-services discoverer))))

; Create a reference from a service
; resolved by CL-ZEROCONF.
; @ INCLUDE EXTENSION FOR GROUP CLASSES
(defun get-service-reference (lambic-record host port)
  (let ((reference (make-instance (if (equal (get-record-value lambic-record
                                                               "group-classp")
                                             "T")
                                      'group-far-reference
                                    'far-reference)
                                  :address host
                                  :port port
                                  :object-name (get-record-value lambic-record
                                                                 "object-name")
                                  :actor-name (get-record-value lambic-record
                                                                "actor-name"))))
    (update-instance-after-reading reference)))

; Store service resolved by CL-ZEROCONF
(defmethod store-resolved-service (discoverer resolved-service)
  (multiple-value-bind (service-description lambic-record)
      (parse-service-description (dns-sd:service-txt-record resolved-service))
    (let ((service-reference (get-service-reference lambic-record
                                                    (dns-sd:service-host resolved-service)
                                                    (dns-sd:service-port resolved-service))))
      (unless (find-imported-service discoverer service-reference service-description)
        (push (cons service-description service-reference) (imported-services discoverer))
        (when (find-required-service discoverer service-description)
          (resolve-service discoverer service-reference service-description)
          (remove-required-description-if-applicable discoverer service-description))))))

; Resolve remote service (far-reference)
(defmethod resolve-service (discoverer service-reference service-description)
  (in-actor-of *communication-actor*
               (create-channel *communication-actor* service-reference
                               service-description)))

; Resolve remote service (far-reference)
(defmethod resolve-service (discoverer (service-reference far-reference)
                                       service-description)
  (when-resolved (call-next-method)
    (lambda (channel)
      (notify-any-discovery-observers discoverer channel service-description
                                      (whenever-observers discoverer)))))

; Resolve remote service (group-far-reference)
(defmethod resolve-service (discoverer (service-reference group-far-reference)
                                       service-description)
  (when-resolved (in-actor-of *communication-actor*
                              (create-channel *communication-actor*
                                              service-reference
                                              service-description))
    (lambda (channels)
      (notify-any-discovery-observers discoverer (car channels)
                                      service-description
                                      (whenever-observers discoverer))
      (notify-any-discovery-observers discoverer (cdr channels)
                                      service-description
                                      (group-whenever-observers discoverer)))))

; Resolve local service (distributable-object)
(defmethod resolve-service (discoverer (service-reference distributable-object)
                                       service-description)
  (notify-any-discovery-observers discoverer service-reference service-description
                                  (whenever-observers discoverer)))

; Import a service with the description passed as parameter.
; The description is stored in the required-services list
; if either the service has not been imported already, or
; if it is being required using the whenever-discovered
; construct.
; This method does NOT return a channel.
(defmethod import-service (discoverer service-description cardinality)
  (let ((service-already-imported nil))
    (dolist (stored-service (imported-services discoverer))
      (when (description-record-equal service-description (car stored-service))
        (setf service-already-imported t)
        (resolve-service discoverer (cdr stored-service) service-description)))
    (when (and (not (find-required-service discoverer service-description))
               (or (not service-already-imported) (equal cardinality :any)))
      (push (cons service-description cardinality)
            (required-services discoverer)))))

; Notify any-discovery observers.
(defmethod notify-any-discovery-observers (discoverer service-channel service-description
                                                      any-observers)
  (loop for observer-cons in (reverse any-observers)
        when (description-record-equal (car observer-cons) service-description)
        do (in-actor-of (cdr observer-cons)
                        (resolve-with-result (cdr observer-cons) service-channel)
                        :with-future nil)))

; Add whenever observer to discoverer
(defmethod add-whenever-observer (discoverer observer service-description)
  (push (cons service-description observer)
        (whenever-observers discoverer))
  (import-service discoverer service-description :any))

; Add group whenever observer to discoverer
(defmethod add-group-whenever-observer (discoverer observer service-description)
  (push (cons service-description observer)
        (group-whenever-observers discoverer))
  (import-service discoverer service-description :any))

; Export object with a user-defined service description
(defmethod export-object-with-description (discoverer object service-description)
  (let* ((object-name (object-name object))
         (actor-name (object-name (object-actor object)))
         (record (append (list (cons "object-name" object-name)
                               (cons "actor-name" actor-name)
                               (cons "group-classp" 
                                     (format nil "~d" (typep object 'group-class))))
                         service-description))
         (dns-sd-service (make-instance 'dns-sd:service
                                        :name object-name
                                        :type "_lambic._tcp"
                                        :host *device-address*
                                        :port *device-port*
                                        :txt-record (dns-sd:build-txt-record record))))
    (push (cons service-description dns-sd-service)
          (published-services discoverer))
    (dns-sd:publish dns-sd-service nil)))

; Publish already exported services.
(defmethod republish-services (discoverer)
  (dolist (service-cons (published-services discoverer))
    (dns-sd:publish (cdr service-cons) nil)))

;;;;;;;;;;;;;;;;;;;;;
; DISCOVERY HANDLER ;
;;;;;;;;;;;;;;;;;;;;;
; Start a process (not an actor) that exclusively handles discovery events.
(defun start-discovery (discoverer)
  (setf *cl-zeroconf-browser*
        (dns-sd:browse "_lambic._tcp" "" (make-instance 'cl-zeroconf-browser)))
  (republish-services discoverer)
  (mp:process-run-function
   "Lambic-zeroconf-handler-process"
   '()
   #'(lambda ()
       (loop
        (unless (boundp '*cl-zeroconf-browser*) (return))
        (sleep 1)
        (dns-sd:process-dns-sd-events 2.0)))))

;;;;;;;;;;;;;;;;;;;;;;;;;
; CL-ZEROCONF callbacks ;
;;;;;;;;;;;;;;;;;;;;;;;;;

; When CL-ZEROCONF discovers a service
; it is asked to resolve such service.
(defmethod dns-sd:browse-add-service (cl-zeroconf-browser
                                      remote-service interface-index more-coming-p)
  (declare (ignore more-coming-p))
  (dns-sd:resolve remote-service cl-zeroconf-browser))

; Resolve a discovered service.
; Invoke Lambic callback with a service description
(defmethod dns-sd:browse-resolved-service (cl-zeroconf-browser remote-service)
  (in-actor-of *discovery-actor*
               (store-resolved-service *discovery-actor* remote-service)
               :with-future nil))

; Notify discovery error
(defmethod dns-sd:browse-error (discoverer error)
  (format t "~&The following error occurred while browsing: ~s~%" error))

(defmethod dns-sd:browse-remove-service (cl-zeroconf-browser service interface-index more-coming-p)
  (declare (ignore more-coming-p))
  "Dummy answer to useless method definition forced by ZEROCONF")


;;;;;;;;;;;;;;;;;;;;;;;
; DISCOVERY FUNCTIONS ;
;;;;;;;;;;;;;;;;;;;;;;;
; Publish a service
(defun publish-service (object service-description)
  (in-actor-of *discovery-actor*
               (export-object-with-description *discovery-actor* object
                                               service-description)
               :with-future nil))

; Install a observer for discovery of multiple services
(defmethod group-whenever-discovered (service-description continuation-lambda)
  (let ((multiple-discovery-future (make-instance 'future)))
    (when-resolved multiple-discovery-future continuation-lambda)
    (in-actor-of *discovery-actor*
                 (add-group-whenever-observer *discovery-actor*
                                              multiple-discovery-future
                                              service-description)
                 :with-future nil)))

; Extension to STANDARD-WHENEVER-DISCOVERED
; Add support for channel groups (used in group classes).
; @ INCLUDE EXTENSION FOR GROUP CLASSES
; Invoke discovery lambda only for the first channel of the group to be
; discovered.
(defmethod whenever-discovered (service-description continuation-lambda)
  (let* ((multiple-discovery-future (make-instance 'future)) 
         discovered-references
         (new-lambda
          (lambda (new-reference)
            (if (or (typep new-reference 'group-channel)
                    (typep new-reference 'group-class))
                (if discovered-references
                    (let ((futures (loop for reference in discovered-references
                                         collect (in-actor-of reference
                                                              (is-member reference
                                                                         new-reference))))
                          member-storedp)
                      (when-any-resolved futures
                        (lambda (memberp lastp)
                          (unless member-storedp
                            (if (not memberp)
                                (when lastp
                                  (push new-reference discovered-references)
                                  (funcall continuation-lambda new-reference))
                              (setf member-storedp t))))))
                  (progn
                    (push new-reference discovered-references)
                    (funcall continuation-lambda new-reference)))
              (funcall continuation-lambda new-reference)))))
    (when-resolved multiple-discovery-future new-lambda)
    (in-actor-of *discovery-actor*
                 (add-whenever-observer *discovery-actor*
                                        multiple-discovery-future
                                        service-description)
                 :with-future nil)))
           
; INITIALISE *discovery-actor*
(setf *discovery-actor* (make-instance 'discoverer))
(spawn-actor "discoverer" *discovery-actor*)
(start-discovery *discovery-actor*)
