;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  (__)  LAMBIC Project
;   )(   Written by Jorge Vallejos
;        
; (c) Vrije Universiteit Brussel, 2008 - 2010
; 
; Permission is hereby granted, free of charge, to any person
; obtaining a copy of this software and associated documentation
; files (the "Software"), to deal in the Software without
; restriction, including without limitation the rights to use,
; copy, modify, merge, publish, distribute, sublicense, and/or
; sell copies of the Software, and to permit persons to whom the
; Software is furnished to do so, subject to the following
; conditions:
;
; The above copyright notice and this permission notice shall be
; included in all copies or substantial portions of the Software.
;
; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
; EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
; OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
; NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
; HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
; WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
; OTHER DEALINGS IN THE SOFTWARE.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; FUTURES IN LAMBIC
; 
; All continuations are executed locally to the sender of the
; message (same host, same actor).
; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package :lambic)

; Continuation lambda to be executed when
; the future is resolved with the result.
(defstruct continuation future lambda)

;;;;;;;;;;;;;;;;
; FUTURE class ;
;;;;;;;;;;;;;;;;
(define-distributable-class future ()
  ((future-state :initarg :state :initform 'UNRESOLVED :accessor future-state)
   (future-result :initarg :result :accessor future-result)
   (future-continuations :initform nil :accessor future-continuations)))

; WHEN-RESOLVED macro
(defmacro when-resolved (future continuation &rest conditions)
  (let ((handled-conditions (remove-if #'(lambda (element) (eql element :catch))
                                       conditions)))
    `(progn
       (when-resolvedf
        (handler-case 
            ,future
          ,@handled-conditions)
        (lambda (result)        
          (if (typep result 'condition)
              (handler-case (error result)
                ,@handled-conditions)
            (funcall ,continuation result))))
       t)))

; The default behaviour of when-resolved method
; is to execute the continuation lambda
; with future's result.
(defmethod when-resolvedf ((this-future future) continuation)
  (funcall continuation (future-result this-future)))

; If the future's result is unbound the continuation
; is stored in the future.
(defmethod when-resolvedf :around ((this-future future) continuation)
  (if (slot-boundp this-future 'future-result)
      (call-next-method)
    (push continuation (future-continuations this-future))))

; Resolve the future with a value as result.
(defmethod resolve-with-result ((resolved-future future) result
                                &key (status 'RESOLVED))
  (setf (future-state resolved-future) status)
  (setf (future-result resolved-future) result)
  (loop for continuation in (reverse (future-continuations resolved-future))
        do (funcall continuation result)))


; Resolve the future with another future as result.
; The future is resolved only when the future passed as result is resolved.
; (future pipelining)
(defmethod resolve-with-result ((resolved-future future) (result future)
                                &key (status 'RESOLVED))
  (when-resolved result (lambda (real-result)
                                (resolve-with-result resolved-future
                                                     real-result :status status))
                 :catch (error (e) (ruin-with-condition resolved-future e))))

; This method is used to resolve a "discardable" remote function invocation
; (indicated with a discard-with-value keyword parameter). This method is called
; when Lambic fails to deliver the remote invocation.
(defmethod resolve-with-discard-result ((resolved-future future) discard-result)
  (resolve-with-result resolved-future discard-result :status 'DISCARDED))

; Ruin a future with a condition.
(defmethod ruin-with-condition ((ruined-future future) condition)
  (setf (future-state ruined-future) 'RUINED)
  (setf (future-result ruined-future) condition)
  (loop for continuation in (reverse (future-continuations ruined-future))
        do (funcall continuation condition)))

; Ruin a future with a condition passed as string.
; It first converts the string into the corresponding condition.
; TODO: This should be improved by using serialisation techniques.
(defmethod ruin-with-condition ((ruined-future future) (condition-text string))
  (ruin-with-condition ruined-future
                       (make-instance 'remote-condition :text condition-text)))

; Boolean to determine if a future has been resolved.
(defmethod resolvedp ((this future))
  (equal (future-state this) 'RESOLVED))

; TIMING METHODS

(defvar *timer-id* 0)

; Ruin future with timeout condition.
(defmethod raise-timeout-condition ((timed-future future))
  (unless (slot-boundp timed-future 'future-result)
    (ruin-with-condition timed-future
                         (make-instance 'timeout-condition
                                        :text (format nil "Future expired: ~D"
                                                      timed-future)))))

;;;;;;;;;;;;;;;;;;;
; Running Example ;
;;;;;;;;;;;;;;;;;;;

#|

; Test 1 (future pipelining)
; Resolve a future (fut2) with another future (fut1):
; Result of fut1 is also the result of fut2.
(defvar fut1 (make-instance 'future))
(defvar fut2 (make-instance 'future))
(when-resolved fut2
               (lambda (result)
                 (format t "final result: ~a" result)))

#|
(resolve-with-result fut2 fut1)
(resolve-with-result fut1 "phew")
|#

; Test 2
; when-resolved returns a future:
; It is resolved with the result of the execution
; of its continuation lambda.
;(defvar fut3 (make-instance 'future))
;(when-resolved (when-resolved fut3
;                              (lambda (result)
;                                (format t "inner result: ~a~%" result)))
;               (lambda (result)
;                 (format t "outer result: ~a~%" result)))

#|
(resolve-with-result fut1 "phew")
|#

; Test 3
; Ruin future with a condition
; After a when-resolved observer has been installed.
(defvar ruinfut1 (make-instance 'future))
(when-resolved ruinfut1
               (lambda (result)
                 (format t "final result: ~a" result)))

#|
(raise-timeout-condition ruinfut1)
|#

; Test 4
; Ruin future with a condition
; Before any when-resolved observer is installed.
(defvar ruinfut2 (make-instance 'future))
(raise-timeout-condition ruinfut2)

#|
(when-resolved ruinfut2
               (lambda (result)
                 (format t "final result: ~a" result)))
|#

; Test 5
; Ruin future with a timeout condition after X seconds.
(defvar ruinfut3 (make-instance 'future))
(when-resolved ruinfut3
               (lambda (result)
                 (format t "final result: ~a" result)))

#|
(timeout-after ruinfut3 10)
|#


|#