;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  (__)  LAMBIC Project
;   )(   Written by Jorge Vallejos
;        
; (c) Vrije Universiteit Brussel, 2008 - 2010
;
; Permission is hereby granted, free of charge, to any person
; obtaining a copy of this software and associated documentation
; files (the "Software"), to deal in the Software without
; restriction, including without limitation the rights to use,
; copy, modify, merge, publish, distribute, sublicense, and/or
; sell copies of the Software, and to permit persons to whom the
; Software is furnished to do so, subject to the following
; conditions:
;
; The above copyright notice and this permission notice shall be
; included in all copies or substantial portions of the Software.
;
; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
; EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
; OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
; NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
; HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
; WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
; OTHER DEALINGS IN THE SOFTWARE.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; LAMBIC MACROS FOR SEQUENTIALITY, CONTEXT PREDICATES
; AND PROPAGATION
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package :lambic)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; DEFINE-PEER-GENERIC macro ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmacro define-peer-generic (function-name (&rest lambda-list)
                                             &body options)
  (multiple-value-bind (updated-options predicates)
      (update-peer-function-options options)
    (let ((updated-lambda-list (update-peer-function-lambda-list lambda-list)))
      `(progn
         (defgeneric ,function-name (,@updated-lambda-list) 
           ,@(when updated-options `(,@updated-options)))
         ,@(when predicates
             `((setf (function-predicates (fdefinition ',function-name))
                     ',predicates)))))))

; Update peer generic function options
(defun update-peer-function-options (options)
  (multiple-value-bind (other-options predicates)
      (extract-predicates-option options)
    (values (ensure-peer-function-classes other-options)
            (parse-function-predicates predicates))))

; Make generic function and its methods use lambic default classes
; (FUTURISED-FUNCTION and PREDICATED-METHOD respectively).
; If function or method classes are provided, keep them.
; TODO: To improve this.
(defun ensure-peer-function-classes (options)
  (let ((function-class (find-if #'(lambda (option)
                                      (eql (car option) :generic-function-class))
                                 options))
        (method-class (find-if #'(lambda (option)
                                      (eql (car option) :method-class))
                                 options)))
    (unless function-class
      (push '(:generic-function-class futurised-function) options))
    (unless method-class
      (push '(:method-class predicated-method) options))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; DEFINE-PEER-METHOD macro ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmacro define-peer-method (&rest forms)
  (multiple-value-bind (name qualifiers specialised-lambda-list declarations
                             documentation body argument-names specialisers
                             method-predicates)
      (parse-peer-method-definition forms)
    (let* ((updated-specialisers (append specialisers
                                         (get-context-specialisers name method-predicates)))
           (updated-lambda-list (update-specialised-lambda-list specialised-lambda-list
                                                                updated-specialisers)))
      
    `(prog2
         (when (or (not (fboundp ',name))
                   (and (fboundp ',name)
                        (not (typep (fdefinition ',name) 'predicated-function))))
           (define-peer-generic ,name (,@argument-names)))
         (defmethod ,name ,@(when qualifiers `(,@qualifiers)) ,updated-lambda-list
           ,@(when declarations `(,@declarations)) ,@(when documentation `(,documentation))
           (futurised-progn ,@body))
       ,@(when method-predicates
           `((let ((method (find-method-with-specialisers (fdefinition ',name)
                                                          ',updated-specialisers)))
               (setf (method-predicates method) ',method-predicates))))))))

; Pare peer method definition
(defun parse-peer-method-definition (forms)
  (multiple-value-bind (name qualifiers specialised-lambda-list declarations
                             documentation body argument-names specialisers)
      (parse-method-definition forms)
    (multiple-value-bind (body predicates)
        (extract-context-predicates body)
      (values name qualifiers specialised-lambda-list declarations
              documentation body argument-names specialisers predicates))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;
; DEFINE-PEER-CLASS macro ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;
#|
(defmacro define-peer-class (class-name superclasses slots &rest class-options)
  (multiple-value-bind (updated-slots accessor-conses)
      (update-slots slots)
    `(progn
       ,@(ensure-futurised-accessors accessor-conses)
       (define-distributable-class ,class-name ,superclasses ,updated-slots ,@class-options))))
|#

(defmacro define-peer-class (class-name superclasses slots &rest class-options)
  (multiple-value-bind (updated-slots accessor-conses)
      (update-slots slots)
    `(progn
       (define-distributable-class ,class-name ,superclasses ,updated-slots ,@class-options)
       ,@(ensure-futurised-accessors accessor-conses class-name))))
