;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  (__)  LAMBIC Project
;   )(   Written by Jorge Vallejos and Pascal Costanza
;        
; (c) Vrije Universiteit Brussel, 2008 - 2010
; 
; Permission is hereby granted, free of charge, to any person
; obtaining a copy of this software and associated documentation
; files (the "Software"), to deal in the Software without
; restriction, including without limitation the rights to use,
; copy, modify, merge, publish, distribute, sublicense, and/or
; sell copies of the Software, and to permit persons to whom the
; Software is furnished to do so, subject to the following
; conditions:
;
; The above copyright notice and this permission notice shall be
; included in all copies or substantial portions of the Software.
;
; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
; EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
; OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
; NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
; HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
; WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
; OTHER DEALINGS IN THE SOFTWARE.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; PASS-BY-REFERENCES SEMANTICS IN LAMBIC
; 
; 
; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package :lambic)

; Global variables
(defvar *device-address* (format nil "~D." (machine-instance)))
(defvar *device-port* 10244)  
(defvar *object-id* 0)
(defvar *distributable-objects* (make-hash-table :test #'equal))
(defvar *far-references* (make-hash-table :test #'equal))
(defvar *screen-output* *standard-output*)
(setf *trace-output* *screen-output*)

; *current-actor* is initialised in the actors.lisp file.
(defvar *current-actor* nil)

(defun debug-print (text object)
  (format *screen-output* "DEBUG. ~D: ~D~%" text object))

(defun generate-new-object-name ()
  (format nil "object-~d-at-~D" 
          (incf *object-id*) *device-address*))

;;;;;;;;;;;;;;;;;;;;;;;
; FAR-REFERENCE class ;
;;;;;;;;;;;;;;;;;;;;;;;
(defclass far-reference (value-object)
  ((reference-address :initarg :address :reader reference-address)
   (reference-port :initarg :port :reader reference-port)
   (reference-actor-name :initarg :actor-name :reader reference-actor-name)
   (reference-object-name :initarg :object-name :reader reference-object-name)))

; @ INCLUDE EXTENSION FOR GROUP CLASSES
(defclass group-far-reference (far-reference) ())

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; DISTRIBUTABLE-OBJECT class ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defclass distributable-object ()
  ((object-name :initarg :object-name :initform (generate-new-object-name)
                :reader object-name)
   (object-actor :initarg :actor :initform *current-actor*
                 :accessor object-actor)))

(defmethod initialize-instance :after ((object distributable-object) &key &allow-other-keys)
  (setf (gethash (object-name object) *distributable-objects*) object))

(defun find-distributable-object (name &optional (errorp t))
  (multiple-value-bind
      (object foundp)
      (gethash name *distributable-objects*)
    (cond (foundp object)
          (errorp (error "Remote object named ~S not found." name)))))

(defun (setf find-distributable-object) (object name &optional errorp)
  (declare (ignore errorp))
  (setf (gethash name *distributable-objects*) object))

(defun make-far-reference (object)
  (make-instance
   'far-reference
   :address *device-address*
   :port *device-port*
   :actor-name (object-name (object-actor object))
   :object-name (object-name object)))

(defun ensure-far-reference (object)
  (or (gethash object *far-references*)
      (setf (gethash object *far-references*)
            (make-far-reference object))))

(defun print-distributable-object (object stream)
  (print (ensure-far-reference object) stream))

(defmethod print-object ((object distributable-object) stream)
  (if *print-readably*
      (print-distributable-object object stream)
    (call-next-method)))

(defmethod update-instance-after-reading ((reference far-reference))
  (if (on-this-machine-p reference)
      (find-distributable-object (reference-object-name reference))
    reference))

(defmethod on-this-machine-p ((reference far-reference))
  (equal (reference-address reference) *device-address*))

(defmethod reference-equal ((reference-1 far-reference) (reference-2 far-reference))
  (and
   (equal (reference-address reference-1) (reference-address reference-2))
   (equal (reference-port reference-1) (reference-port reference-2))
   (equal (reference-actor-name reference-1) (reference-actor-name reference-2))
   (equal (reference-object-name reference-1) (reference-object-name reference-2))))

; DEFINE-DISTRIBUTABLE-CLASS macro
; Make distributable-object a superclass of the defined class.
(defmacro define-distributable-class (name (&rest superclass-names)
                                           slots &rest class-options)
  (let* ((superclasses (find-classes superclass-names))
         (d-o-class (find-class 'distributable-object))
         (superclass-names (find-names (if (find-superclass d-o-class superclasses)
                                           superclasses
                                         (push d-o-class superclasses)))))
    `(defclass ,name (,@superclass-names) ,slots ,@class-options)))

; DEFPROTOTYPECLASS macro
(defmacro defprotoclass (name (&rest superclass-names)
                              slots &rest class-options)
  (let* ((superclasses (find-classes superclass-names))
         (d-o-class (find-class 'value-object))
         (superclass-names (find-names (if (find-superclass d-o-class superclasses)
                                           superclasses
                                         (push d-o-class superclasses)))))
    `(defclass ,name (,@superclass-names) ,slots ,@class-options)))