;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  (__)  LAMBIC Project
;   )(   Written by Pascal Costanza and Jorge Vallejos
;        
; (c) Vrije Universiteit Brussel, 2008 - 2010
; 
; Permission is hereby granted, free of charge, to any person
; obtaining a copy of this software and associated documentation
; files (the "Software"), to deal in the Software without
; restriction, including without limitation the rights to use,
; copy, modify, merge, publish, distribute, sublicense, and/or
; sell copies of the Software, and to permit persons to whom the
; Software is furnished to do so, subject to the following
; conditions:
;
; The above copyright notice and this permission notice shall be
; included in all copies or substantial portions of the Software.
;
; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
; EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
; OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
; NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
; HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
; WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
; OTHER DEALINGS IN THE SOFTWARE.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; PASS-BY-REFERENCE SEMANTICS IN LAMBIC
; 
; 
; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package :lambic)

(defgeneric update-instance-after-reading (object)
  (:method ((object t)) object))

;;;;;;;;;;;;;;;;
; VALUE-OBJECT ;
;;;;;;;;;;;;;;;;

(defclass value-object () ())

(defun value-object (class-name &rest initargs)
  (declare (dynamic-extent initargs))
  (let ((object (allocate-instance (find-class class-name))))
    (loop for (slot-name value) on initargs by #'cddr do
          (setf (slot-value object slot-name) value))
    (update-instance-after-reading object)))

(defun print-value-object (object stream)
  (let ((*print-circle* nil))
    (format stream "#.(~S '~S" 'value-object (class-name (class-of object)))
    (loop for slot in (class-slots (class-of object))
          unless (eq (slot-definition-allocation slot) :class) do
          (let* ((slot-name (slot-definition-name slot))
                 (slot-value (slot-value object slot-name)))
            (when (typep slot-value 'sequence)
              (setf slot-value (copy-seq slot-value)))
            (format stream " '~S '~S" slot-name slot-value)))
    (format stream ")")))

(defmethod print-object ((object value-object) stream)
  (if *print-readably*
    (print-value-object object stream)
    (call-next-method)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; FUNCALLABLE-VALUE-OBJECT ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass funcallable-value-object ()
  ((definition :initarg :definition :reader funcallable-instance-definition))
  (:metaclass funcallable-standard-class))

(defmethod initialize-instance :after
  ((object funcallable-value-object) &key definition)
  (closer-common-lisp-user::set-funcallable-instance-function object (compile nil definition)))

(defmethod reinitialize-instance :after
  ((object funcallable-value-object) &key (definition nil definitionp))
  (when definitionp
    (closer-common-lisp-user::set-funcallable-instance-function object (compile nil definition))))

(defmethod update-instance-after-reading :after ((object funcallable-value-object))
  (closer-common-lisp-user::set-funcallable-instance-function object (compile nil (funcallable-instance-definition object))))

(defmethod print-object ((object funcallable-value-object) stream)
  (if *print-readably*
    (print-value-object object stream)
    (call-next-method)))

(defmacro vlambda ((&rest lambda-list) &body body)
  `(make-instance 'funcallable-value-object
                  :definition '(lambda ,lambda-list ,@body)))

(defmacro define-vfunction (name (&rest lambda-list) &body body)
  `(progn
     (declaim (notinline ,name))
     (defun ,name ,lambda-list)
     (setf (fdefinition ',name)
           (vlambda ,lambda-list ,@body))
     ',name))
