;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  (__)  LAMBIC Project
;   )(   Written by Jorge Vallejos
;        
; (c) Vrije Universiteit Brussel, 2008 - 2010
; 
; Permission is hereby granted, free of charge, to any person
; obtaining a copy of this software and associated documentation
; files (the "Software"), to deal in the Software without
; restriction, including without limitation the rights to use,
; copy, modify, merge, publish, distribute, sublicense, and/or
; sell copies of the Software, and to permit persons to whom the
; Software is furnished to do so, subject to the following
; conditions:
;
; The above copyright notice and this permission notice shall be
; included in all copies or substantial portions of the Software.
;
; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
; EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
; OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
; NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
; HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
; WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
; OTHER DEALINGS IN THE SOFTWARE.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; LAMBIC PROPAGATION
;
; MOP extension to enable:
;
; - group classes 
; - group methods
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package :lambic)

;;;;;;;;;;;;;;;;;;;;;;;;
; DEFGROUPCLASS macro ;
;;;;;;;;;;;;;;;;;;;;;;;;

; Define a subclass of GROUP-CLASS with an identity function.
(defmacro defgroupclass (class-name superclasses slots &rest class-options)
    (multiple-value-bind (new-class-options identity-function)
      (extract-identity-function class-options)
      (let* ((group-subclassp (find-superclass (find-class 'group-class)
                                               (find-classes superclasses)))
             (new-superclasses (append (unless group-subclassp
                                         '(group-class))
                                       superclasses)))
        `(progn
           (define-peer-class ,class-name ,new-superclasses
                              ,slots ,@new-class-options)
           ,(unless group-subclassp
              (generate-initialize-instance class-name identity-function))
           (find-class ',class-name)))))

; Generate group lookup method
(defun generate-initialize-instance (class-name identity-function)
  `(c2cl:defmethod initialize-instance ((object ,class-name)
                                          &key &allow-other-keys)
     (let* ((object (call-next-method))
            (description (list (cons "class-name" (format nil "~d" ',class-name)))))
       ,@(when identity-function
           `((setf (identity-function object) ,identity-function)))
       (add-peer object object)
       (publish-service object description)
       (group-whenever-discovered description
         (lambda (peer)
           (when (and (not (eql peer object))
                      (not (find-peer peer (all-peers object))))
             (when-resolved (is-peer object peer)
               (lambda (memberp)
                 (when memberp
                   (add-peer object peer)
                   (peer-discovered object peer)
                   (when-disconnected peer (lambda (peer) 
                                             (peer-disconnected object peer)))
                   (when-reconnected peer (lambda (peer) 
                                            (peer-reconnected object peer)))
                   (when-self-disconnected object
                                           (lambda (self)
                                             (local-disconnection self)))
                   (when-self-reconnected object
                                          (lambda (self)
                                            (local-reconnection self))))))))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; AUXILIARS FOR DEFGROUPCLASS ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; Extract identity function option
(defun extract-identity-function (options)
  (multiple-value-bind (options identity-property)
      (remove-if* #'(lambda (option) (eql (car option) :identity-function))
                  options :return-removed-items t)
    (let ((id-lambda (cadar identity-property)))
      (when (and id-lambda (typep id-lambda 'list))
        (multiple-value-bind (lambda-list declarations
                                          documentation body)
            (parse-lambda id-lambda)
          `(futurised-lambda ,lambda-list
             ,@(when declarations `(,@declarations))
             ,@(when documentation `(,documentation))
             ,@body)))
      (values options id-lambda))))

; Default implementation of PEER-DISCOVERED event handler
(define-peer-method peer-discovered ((object group-class) peer)
  (format *screen-output* "Local IM ~d has discovered ~d~%" object peer))

; Default implementation of PEER-DISCONNECTED event handler
(define-peer-method peer-disconnected ((object group-class) peer)
  (format *screen-output* "~d has disconnected from Local IM ~d~%" object peer))

; Default implementation of PEER-RECONNECTED event handler
(define-peer-method peer-reconnected ((object group-class) peer)
  (format *screen-output* "~d has reconnected with Local IM ~d~%" object peer))

; Default implementation of LOCAL-DISCONNECTION event handler
(define-peer-method local-disconnection ((object group-class))
  (debug-print "Default Local disconnection handler for" object))

; Default implementation of LOCAL-RECONNECTION event handler
(define-peer-method local-reconnection ((object group-class))
  (debug-print "Default Local reconnection handler for" object))

; Check identity function
(define-peer-method is-peer ((member group-class) aspirant)
  (futurised-funcall (identity-function member) member aspirant))

; Get peers of object's group
(defun group-peers (object &optional (class (class-of object)))
  (loop for peer in (all-peers object)
        when (or (typep peer (class-name class))
                 (typep peer 'generic-channel))
        collect peer))

; Add member
(define-peer-method add-peer ((object group-class) peer)
  (push peer (all-peers object)))

; Find member (for far reference)
(defmethod find-peer ((peer far-reference) peers)
  (find-if #'(lambda (member)
               (peer-equal member peer))
           peers))

; Find member
(defmethod find-peer (peer peers)
  (find peer peers))

(defmethod peer-position (peer peers)
  (position peer peers))

(defmethod peer-position ((peer-ref far-reference) peers)
  (let ((channel (find-peer peer-ref peers)))
    (call-next-method channel peers)))

(defmethod peer-equal ((peer-1 far-reference) (peer-2 group-class))
   (equal (reference-object-name peer-1) (object-name peer-2))) 

(defmethod peer-equal ((peer-1 group-class) (peer-2 far-reference))
   (equal (object-name peer-1) (reference-object-name peer-2))) 

(defmethod peer-equal ((peer-1 far-reference) (peer-2 far-reference))
  (reference-equal peer-1 peer-2))

(defmethod peer-equal ((peer-1 group-class) (peer-2 group-class))
   (eql peer-1 peer-2))

(defmethod peer-equal ((peer-1 generic-channel) (peer-2 generic-channel))
  (eql peer-1 peer-2))

(defmethod peer-equal ((peer-1 far-reference) (peer-2 generic-channel))
  (reference-equal peer-1 (channel-reference peer-2)))

(defmethod peer-equal ((peer-1 generic-channel) (peer-2 far-reference))
  (reference-equal (channel-reference peer-1) peer-2))

(defmethod peer-equal (peer-1 peer-2)
  nil)

; Default publishing record
(define-peer-method publish-with-record ((object group-class))
  nil)


; EXTENSION OF DISCOVERY SERVICE
; Resolve local service (group-class)
(defmethod resolve-service (discoverer (service-reference group-class)
                                       service-description)
  (notify-any-discovery-observers discoverer service-reference service-description
                                  (whenever-observers discoverer))
  (notify-any-discovery-observers discoverer service-reference service-description
                                  (group-whenever-observers discoverer)))

; Default identity function: same class
; This is always T as the class is one of the conditions to
; discover the peer.
(defun same-class (local-peer remote-peer)
  (declare (ignore local-peer remote-peer))
  t)


;;;;;;;;;;;;;;;;;;;;;;;;;
; DEFGROUPGENERIC macro ;
;;;;;;;;;;;;;;;;;;;;;;;;;
(defmacro defgroupgeneric (function-name (&rest lambda-list)
                                             &body options)
  (multiple-value-bind (group-options peer-options)
      (update-group-function-options options)  
    (let ((group-function-name (get-group-function-name function-name)))
      `(progn
         (define-peer-generic ,group-function-name ,lambda-list ,@group-options)
         (prog1
             (define-peer-generic ,function-name ,lambda-list ,@peer-options)
           (let ((new-group-function (fdefinition ',group-function-name))
                 (new-peer-function (fdefinition ',function-name)))
             (setf (peer-group-function new-peer-function) new-group-function)
             (define-peer-method ,group-function-name ,lambda-list
               (let ((*group-level-call* t)
                     (*original-arguments* (list ,@lambda-list)))
                 ;(format *screen-output* "*original-arguments* in base group method: ~d~%" *original-arguments*)
                 (funcall (fdefinition ',function-name) ,@lambda-list)))))))))

; Update group generic function options. Return the options for the
; peer an group generic functions.
(defun update-group-function-options (options)
  (multiple-value-bind (options peer-predicates)
      (remove-if* #'(lambda (option) (eql (car option) :predicates))
                  options :return-removed-items t)
    (multiple-value-bind (options group-predicates)
        (remove-if* #'(lambda (option) (eql (car option) :group-predicates))
                    options :return-removed-items t)
      (setf options
            (remove-if #'(lambda (option) 
                           (eql (car option) :generic-function-class))
                       options))
      (let* ((group-options (append options
                                    (when group-predicates
                                      `((:predicates ,@(cdar group-predicates))))
                                    '((:generic-function-class group-function))))
             (peer-options (append options peer-predicates)))
;             (peer-options (append options (car peer-predicates))))
        (values group-options peer-options)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; AUXILIARS FOR DEFGROUPGENERIC ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; Create group function name
(defun get-group-function-name (name)
  (read-from-string (format nil "group-~d" name)))

; Get the predicates defined for the peer function's group function
(defmethod group-predicates ((function peer-function))
  (function-predicates (peer-group-function function)))

; Set the predicates defined for the peer function's group function
(defmethod set-group-predicates ((function peer-function) predicates)
  (setf (function-predicates (peer-group-function function)) predicates))

;;;;;;;;;;;;;;;;;;;;;;;;
; DEFGROUPMETHOD macro ;
;;;;;;;;;;;;;;;;;;;;;;;;

(defmacro defgroupmethod (&rest forms)
  (multiple-value-bind (name qualifiers lambda-list declarations
                             documentation body argument-names
                             specialisers propagation-plist)
      (parse-group-method-definition forms)
    (let ((group-method-name (read-from-string (format nil "group-~d"
                                                       name)))
          (expanded-body (expand-call-next-method body name argument-names
                                                  specialisers
                                                  propagation-plist)))
      `(define-peer-method ,group-method-name
                           ,@(when qualifiers `(,@qualifiers))
                           ,lambda-list
                           ,@(when declarations `(,@declarations))
                           ,@(when documentation `(,documentation))
                           ,@expanded-body))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; AUXILIARS FOR DEFGROUPMETHOD ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; Parse a group method definition
(defun parse-group-method-definition (forms)
  (multiple-value-bind (name qualifiers lambda-list declarations
                             documentation body argument-names specialisers)
      (parse-method-definition forms)
    (multiple-value-bind (body propagation-plist)
        (remove-if* #'(lambda (expression)
                        (and (listp expression)
                             (eql (car expression) :propagate)))
                    body :return-removed-items t)
      (let ((predicates (find-if #'(lambda (expression)
                                     (when (listp expression) 
                                       (eql (car expression) :when)))
                                 body)))
        (setf specialisers 
              (append specialisers 
                      (get-context-specialisers (read-from-string (format nil "GROUP-~d"
                                                                          name))
                                                (cdr predicates))))
        (values name qualifiers lambda-list declarations documentation
                body argument-names specialisers (car propagation-plist))))))

; Replace CALL-NEXT-METHOD form with the propagation form
(defun expand-call-next-method (body method-name argument-names specialisers
                                     propagation-plist)
  (let* ((conditions (getf propagation-plist :propagate))
         (propagation-function (get-propagation-function method-name specialisers
                                                         argument-names propagation-plist))
         (expansion (call-next-method-expansion conditions
                                                propagation-function
                                                argument-names
                                                specialisers)))
    (expand-form 'call-next-method body expansion)))

(defvar *original-arguments*)
;(defvar *original-argument-names*)

; Generate a propagation form.
; This function returns a lambda as result as CALL-NEXT-METHOD can
; receive parameters which should be considered in the propagation
; form. However, those parameters are ignored so far.
(defun call-next-method-expansion (conditions propagation-function argument-names
                                              specialisers)
  `(let* ((all-combinations (find-argument-combinations (list ,@argument-names)
                                                        ',specialisers))
;          (*original-arguments* (list ,@argument-names))
          (*original-arguments* (list ,@argument-names))
          ;(*original-argument-names* ',argument-names)
          ;(*original-arguments* ,(loop for argument-name in argument-names
          ;                             collect `(cons ',argument-name ,argument-name)))
          (selected-combinations (apply-propagation-conditions ,conditions
                                                               ,argument-names
                                                               all-combinations))
          (call-next-method-future (make-instance 'future)))
     (when-all-resolved (list ,propagation-function selected-combinations)
       (lambda (results)
         (resolve-with-result call-next-method-future
                              (funcall (first results) (second results)))))
     call-next-method-future))
#|
     (futurised-funcall ,propagation-function
                        selected-combinations)))
|#

; Filter argument combinations according to propagation
; conditions.
(defmacro apply-propagation-conditions (conditions lambda-list original-combinations)
  `(flet ((propagation-condition (combination)
            (let ((*original-arguments* (list ,@lambda-list))
;(*original-argument-names* (progn
                  ;                             (debug-print "(list ,@lambda-list)" (list ,@lambda-list))
                  ;                             (list ,@lambda-list)))
                  ,@(when lambda-list
                       (loop for i from 0 to (1- (length lambda-list))
                             collect `(,(nth i lambda-list) (nth ,i combination)))))
              (parallel-and ,@conditions))))
     (let ((selection-future (make-instance 'future)))
       (when-resolved ,original-combinations
         (lambda (original-combinations)
           (let ((test-futures (loop for tested-combination in original-combinations
                                     collect (propagation-condition tested-combination))))
             (when-all-resolved test-futures
               (lambda (tested-combinations)
                 (let ((selected-combinations (loop for tested in tested-combinations
                                                    for original in original-combinations
                                                    when tested
                                                    collect original)))
                 (resolve-with-result selection-future selected-combinations)))))))
             selection-future)))

; Expand a form starting with the symbol passed as parameter
; and using the expander function also passed as parameter.
(defun expand-form (symbol forms expansion)
  (when forms
    (loop for form in forms
          for i from 0 to (1- (length forms))
          do (when (typep form 'list)
               (if (eql (car form) symbol)
                   (setf (nth i forms) expansion)
                 (expand-form symbol form expansion)))))
  forms)

; Find all the combinations of arguments taking into account
; each of the members of the group arguments.
(defun find-argument-combinations (args specialisers)
  (when args
    (let* ((group-class (find-class 'group-class))
           (expanded-args (loop for arg in args
                                for i from 0 to (1- (length args))
                                if (or (eql (nth i specialisers) group-class)
                                       (find-superclass group-class
                                                        (class-direct-superclasses 
                                                         (nth i specialisers))))
                                collect (group-peers arg (nth i specialisers))
                                else
                                collect (list arg))))
      (compose-argument-combinations (car expanded-args)
                                     (cdr expanded-args) nil))))

#|
(defun find-argument-combinations (args specialisers)
  (when args
    (let ((expanded-args (loop for arg in args
                               for i from 0 to (1- (length args))
                               if (typep arg 'group-class)
                               collect (group-peers arg (nth i specialisers))
                               else
                               collect (list arg))))
      (compose-argument-combinations (car expanded-args)
                                     (cdr expanded-args) nil))))
|#

; Compose the combinations of arugments
(defun compose-argument-combinations (first rest row)
  (if rest
      (mappend #'(lambda (el) 
                   (compose-argument-combinations (car rest) (cdr rest)
                                                  (append row
                                                          (list el))))
               first)
    (mapcar #'(lambda (el) (append row (list el))) first)))


; Get propagation function
(defun get-propagation-function (method-name specialisers argument-names propagation-plist)
  (let ((sequential-propagationp (getf propagation-plist :in-sequence))
        (properties (get-propagation-properties propagation-plist argument-names)))
    (if sequential-propagationp
        (sequential-propagation method-name specialisers argument-names properties)
      (parallel-propagation method-name specialisers argument-names properties))))

; Sequential propagation
; Send method invocations using a FUTURISED-DOLIST. Use the handlers
; for timeout exceptions specified in the catch or negligible argument.
; If there is stop condition, break the loop when such condition is
; satisfied (using RETURN). Reduce the result using the reduce-results-to function
; (if any).
(defun sequential-propagation (method-name specialisers argument-names properties)
  (declare (ignore argument-names)) ; Only for the moment
  (let ((stop-condition (getf properties :stop-condition nil))
        (reduce-results-to (getf properties :reduce-results-to 
                                 (fdefinition 'default-reduce-results-to)))
        (timeout (getf properties :timeout *standard-timeout*))
        (negligible (getf properties :negligible nil))
        (catch (getf properties :catch nil)))
  `(futurised-lambda (argument-combinations)
     (let* ((*standard-timeout* ,timeout)
            (results (futurised-dolist (combination argument-combinations results)
                       (let ((result ,(make-method-call method-name specialisers
                                                        `combination negligible catch)))
                         (append results (list result))
                         (futurised-when ,(when stop-condition
                                            `((funcall ,stop-condition result)))
                           (return results))))))
     ,(when reduce-results-to
        `(funcall ,reduce-results-to final-results))))))

; Parallel propagation
; Send method invocations using a simple dolist and store the result
; futures. Use the handlers for timeout exceptions specified in the
; catch or negligible argument. If there is stop condition use
; WHEN-ANY-RESOLVED to check such condition each time a result is
; received. If no stop condition is defined, return the list of
; futures. Reduce the result using the reduce-results-to function (if any).
(defun parallel-propagation (method-name specialisers argument-names properties)
  (let ((stop-condition (getf properties :stop-condition nil))
        (reduce-results-to (getf properties :reduce-results-to 
                                 (fdefinition 'default-reduce-results-to)))
        (timeout (getf properties :timeout *standard-timeout*))
        (negligible (getf properties :negligible nil))
        (catch (getf properties :catch nil)))
    `(futurised-lambda (argument-combinations)
       ;(debug-print "INVOKING combinations" argument-combinations)
       (let* ((parallel-future (make-instance 'future))
              (*standard-timeout* ,timeout)
              (intermediate-results
               (loop for combination in argument-combinations
                     collect ,(make-method-call method-name specialisers `combination
                                                negligible catch)))
              ,@(when (and stop-condition reduce-results-to)
                  `(final-results)))
         ,@(if stop-condition
               `((when-any-resolved intermediate-results
                   (lambda (result lastp)
                     (append final-results (list result))
                     (futurised-when 
                      (futurised-or (funcall ,stop-condition result) lastp)
                      ,@(when reduce-results-to
                          `((resolve-with-result parallel-future
                                                 (funcall ,reduce-results-to
                                                          final-results))))))))
             (when reduce-results-to
               `((when-resolved intermediate-results
                  (lambda (results)
                    (let ((*original-arguments* (list ,@argument-names)))
                      (resolve-with-result parallel-future
                                           (funcall ,reduce-results-to results))))))))
         ,@(when reduce-results-to `(parallel-future))))))

; Make a method call
(defun make-method-call (method-name specialisers combined-arguments
                                     negligible catch)
  (let ((specialiser-names (get-specialiser-names specialisers)))
  `(handler-case
       (if (in-current-actor (car ,combined-arguments))
           (call-next-method ,@(when specialisers
                                 (loop for i from 0 to (1- (length specialisers))
                                       collect `(nth ,i ,combined-arguments))))
         (apply ',method-name (append ,combined-arguments
                                      ,@(when specialisers
                                          `((append (list :method-specialisers) 
                                                    (list ',specialiser-names)))))))
     ,@(if catch
         `(,catch)
         (when negligible `((timeout-condition () nil)))))))

; Get propagation properties
; Remove :PROPAGATE and :IN-SEQUENCE key/value pairs. Futurise
; :STOP-CONDITION and :REDUCE-RESULTS-TO values (lambdas) if any.
(defun get-propagation-properties (propagation-plist argument-names)
  (let* ((stop-condition (get-property-function (getf propagation-plist :stop-condition)
                                                argument-names))
         (reduce-results-to (get-property-function (getf propagation-plist :reduce-results-to)
                                                   argument-names)))
    (when stop-condition
      (setf (getf propagation-plist :stop-condition) stop-condition))
    (when reduce-results-to
      (setf (getf propagation-plist :reduce-results-to) reduce-results-to))
    (remf propagation-plist :propagate)
    (remf propagation-plist :in-sequence)
    propagation-plist))

(defun get-property-function (function-description lambda-list)
  (when function-description
    (lambda (value)
;      (debug-print "*original-arguments*" "in get-property")
;      (debug-print "is" *original-arguments*)
      (restart-case
          (handler-bind ((unbound-variable
                          #'(lambda (c)
                              (let ((position (position (cell-error-name c) lambda-list)))
                                (if position
                                    (invoke-restart 'use-value
                                                    (nth position *original-arguments*))
                                  (error c))))))
            (funcall (if (symbolp function-description)
                         (fdefinition function-description)
                       (eval function-description)) 
                     value))))))

#|
(defun get-propagation-properties (propagation-plist)
  (let* ((stop-condition (eval (getf propagation-plist :stop-condition)))
         (reduce-results-to (eval (getf propagation-plist :reduce-results-to))))
    (when stop-condition
      (setf (getf propagation-plist :stop-condition)
            (futurised-lambda (result) (funcall stop-condition result))))
    (when reduce-results-to
      (setf (getf propagation-plist :reduce-results-to)
            (futurised-lambda (results) (funcall reduce-results-to results))))
    (remf propagation-plist :propagate)
    (remf propagation-plist :in-sequence)
    propagation-plist))
|#

; Reduce-Results-To or stop-condition function
(define-peer-method first-in (results)
  (let ((first-in-future (make-instance 'future)))
    (when-first-resolved results
      (lambda (result)
        (resolve-with-result first-in-future result)))
    first-in-future))

; Propagation predicate
(define-peer-method each (arg) t)

; Default reduce-results-to
(define-peer-method default-reduce-results-to (results) results)

; Propagation predicate
(define-peer-method this (arg)                   
  (find arg *original-arguments*))

; Get original binding for an argument
(define-peer-method get-original-binding (arg)
  (let ((members (all-peers arg)))
;    (debug-print "*original-arguments*" "in get-original-binding:")
;    (debug-print "is" *original-arguments*)
    (loop for member in members
          when (find member *original-arguments*)
          return member)))

(define-peer-method get-original-binding ((channel generic-channel))
  ;(debug-print "*current-actor*" *current-actor*)
  (loop for argument in *original-arguments*
        when (find channel (all-peers argument))
        return argument))

#|(define-peer-method get-original-binding (symbol function-symbol)
  (let* ((function (fdefinition function-symbol))
  (debug-print "get-original-binding for" symbol)
  (debug-print "*original-argument-names*" *original-argument-names*)
  (debug-print "(eval symbol)" (eval symbol))
  (eval symbol))
;  (let ((position (position symbol *original-argument-names*)))
;    (nth position *original-arguments*)))
|#



    

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; PEER AND GROUP FUNCTION EXECUTIONS ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; Compute discriminating function for the PEER-FUNCTION class
; Invoke the group function binding the *CALL-PEER-METHOD* variable
; which is invoked in the default group method definition.
(defmethod compute-discriminating-function ((function peer-function))
  (lambda (&rest args)
    ;(debug-print "peer-function" function)
    (let ((group-function (peer-group-function function)))
      (if (or *group-level-call* (not group-function))
          (let* ((*group-level-call* nil)
                 (result (apply (call-next-method) args)))
            result)
        (apply group-function args)))))

; Compute discriminating function for the GROUP-FUNCTION class.
; In case of propagation, execute only the methods after the last
; method executed in another host. This is known if the args
; has a keyword argument :METHOD-SPECIALISERS which contains the
; the list of specialisers (symbols, not classes) of the last
; executed method.
(defmethod compute-discriminating-function ((function group-function))
  (lambda (&rest args)
    ;(debug-print "group-function" function)
    (multiple-value-bind (original-args specialisers)
        (extract-method-specialisers args)
      (if specialisers
          (let ((group-function-future (make-instance 'future))
                (methods-future (compute-applicable-predicated-methods function 
                                                                       original-args)))
#|
                (methods-future (compute-applicable-group-methods function 
                                                                  original-args)))
|#
            (when-resolved methods-future
              (lambda (selected-methods)
                (resolve-with-result group-function-future
                                     (apply-selected-group-methods selected-methods 
                                                                   function specialisers
                                                                   original-args))))
            group-function-future)
        (apply (call-next-method) original-args)))))

; Compute applicable predicated methods. Used in the method
; compute-discriminating-function of the GROUP-FUNCTION class.
#|
(defmethod compute-applicable-group-methods ((function group-function) args)
  (let* ((result-future (make-instance 'future))
         (predicate-args (make-list (- (length (generic-function-lambda-list function))
                                       (length args))
                                    :initial-element (make-instance 'lambic-predicate-9)))
         (extended-args (append args predicate-args))
         (original-methods (compute-applicable-methods function extended-args))
         (future-list (futurised-remove-if #'(lambda (method)
                                               (eval-method-predicates method args))
                                           original-methods)))
    (when-resolved future-list
      (lambda (selected-methods)
        (resolve-with-result result-future selected-methods)))
    result-future))
|#

; Apply selected predicated methods for a group function.
(defmethod apply-selected-group-methods (methods (function group-function) specialisers args)
  (let* ((predicate-args (make-list (- (length (generic-function-lambda-list function))
                                       (length args))
                                    :initial-element (make-instance 'lambic-predicate-9)))
         (extended-args (append args predicate-args))
         (specialiser-classes (get-specialiser-classes specialisers))
         (remaining-methods (remove-executed-methods methods specialiser-classes)))
    (apply-selected-methods remaining-methods function extended-args)))

(defmethod compute-peer-applicable-methods (function args)
  (multiple-value-bind (peer-args specialisers)
      (extract-method-specialisers args)
    (declare (ignore specialisers))
    (compute-applicable-predicated-methods function peer-args)))
#|
    (let ((methods-future (compute-applicable-predicated-methods function peer-args)))
      (if (not specialisers)
          methods-future
        (let ((specialiser-classes (get-specialiser-classes specialisers))
              (peer-methods-future (make-instance 'future)))
          (when-resolved methods-future
                         (lambda (methods)
                           (resolve-with-result peer-methods-future 
                                                (remove-executed-methods methods
                                                                         specialiser-classes))))
          peer-methods-future)))))
|#

; Extract the method specialisers from the original argument list
(defun extract-method-specialisers (args)
  (let ((key-position (position :method-specialisers args))
        method-specialisers)
    (when key-position
      (setf method-specialisers (car (last args)))
      (setf args (remove :method-specialisers
                         (remove method-specialisers args))))
    (values args method-specialisers)))

; Get the specialisers' classes.
(defun get-specialiser-classes (specialisers)
  (loop for specialiser in specialisers
        if (typep specialiser 'list)
        ; For EQL specialisers
        collect (second specialiser)
        else
        ; For class specialisers
        collect (find-class specialiser)))

; Remove all methods already executed.
(defun remove-executed-methods (methods method-specialisers)
;  (format *screen-output* "~%~%~%~%")
;  (debug-print "METHOD SPECIALISERS" method-specialisers)
;  (format *screen-output* "~%")
;  (debug-print "METHODS" methods)
  (let ((methods (reverse (loop for method in (reverse methods)
                                until (specificity->= (method-specializers method)
                                                      method-specialisers)
                                collect method))))
;    (format *screen-output* "~%")
;    (debug-print "WO-EXECUTED-METHODS" methods)
;    (format *screen-output* "~%~%~%~%")
    methods))
