;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  (__)  LAMBIC Project
;   )(   Written by Jorge Vallejos
;        
; (c) Vrije Universiteit Brussel, 2008 - 2010
; 
; Permission is hereby granted, free of charge, to any person
; obtaining a copy of this software and associated documentation
; files (the "Software"), to deal in the Software without
; restriction, including without limitation the rights to use,
; copy, modify, merge, publish, distribute, sublicense, and/or
; sell copies of the Software, and to permit persons to whom the
; Software is furnished to do so, subject to the following
; conditions:
;
; The above copyright notice and this permission notice shall be
; included in all copies or substantial portions of the Software.
;
; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
; EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
; OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
; NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
; HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
; WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
; OTHER DEALINGS IN THE SOFTWARE.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; LAMBIC FUTURE-BASED SEQUENTIALITY LIBRARY
;
; Futurised versions of Common Lisp special operators, macros,
; and functions (to enable them to deal with futures). Also,
; definition of the FUTURISE-FUNCTIONS macro to automatically
; convert functions.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package :lambic)

;;;;;;;;;;
; MACROS ;
;;;;;;;;;;

; FUTURISE-FUNCTIONS
; IMPORTANT: In order to futurise a function, it should be
; shadowed for the current package.
; e.g. (futurise-functions print (format :from :cl))
; TODO: Could this enable one to write 'print instead of print?
; (prog1 (read-from-string (format nil "~d" 'print)))
(defmacro futurise-functions (descriptors)
  `(progn
     ,@(loop for descriptor in (cadr descriptors)
             collect
             (let* ((name (if (typep descriptor 'list)
                              (first descriptor)
                            descriptor))
                    (package (if (typep descriptor 'list)
                                 (third descriptor)
                               :c2cl))
                    (function (fdefinition (find-symbol name package))))
               `(defun ,name (&rest args)
                  (let ((future (make-instance 'future)))
                    (when-all-resolved args
                      (lambda (results)
                        (resolve-with-result future (apply ,function results))))
                    (if (resolvedp future)
                        (future-result future)
                      future)))))))

; FUTURISED-DOLIST
; Execute the expressions of its body sequentially,
; even in the presence of futures.
(defmacro futurised-dolist ((iterator-name iterator-list
                                           &optional (iterator-result
                                                      nil
                                                      iterator-result-supplied-p))
                            &body body)
  `(when ,iterator-list
     (let (iteration-haltedp
           ,@(when iterator-result-supplied-p
               `(,iterator-result
                 (iterator-future (make-instance 'future)))))
       (macrolet ((return (&optional (what nil what-supplied-p))
                    `(progn
                       (setf iteration-haltedp t)
                       (return-from iteration-block ,@(when what-supplied-p
                                                        `(,what))))))
         (labels ((iterate (future-name iteration-rest)
                    (block iteration-block
                      (when-resolved future-name
                        (lambda (,iterator-name)
                          (when-resolved (futurised-progn ,@body)
                                         (lambda (iteration-result)
                                           (if iteration-haltedp
                                               iteration-result
                                             (when (car iteration-rest)
                                               (iterate (car iteration-rest)
                                                        (cdr iteration-rest)))))))))))
           (when-resolved ,iterator-list
             (lambda (list)
               (when list
                 (iterate (car list) (cdr list))
                 ,@(when iterator-result-supplied-p
                     `((when-all-resolved ,iterator-result
                                          (lambda (results)
                                            (resolve-with-result iterator-future results))))))))
           ,@(when iterator-result-supplied-p
               `((if (resolvedp iterator-future)
                     (future-result iterator-future)
                   iterator-future))))))))

; FUTURISED-IF
(defmacro futurised-if (&rest args)
  `(let ((fut (make-instance 'future))
         (condition ,(car args)))
     (when-resolved condition
                    (lambda (condition-result)
                      (resolve-with-result fut (if condition-result ,@(cdr args)))))
     (if (resolvedp fut)
         (future-result fut)
       fut)))

; FUTURISED-WHEN
(defmacro futurised-when (&rest args)
  `(futurised-if ,(car args)
                 (futurised-progn ,@(cdr args))))

; FUTURISED-UNLESS
(defmacro futurised-unless (&rest args)
  `(futurised-if ,(car args)
                 nil
                 (futurised-progn ,@(cdr args))))

; FUTURISED-HANDLER-CASE
(defmacro futurised-handler-case (expression &rest cases)
  `(let ((hcfuture (make-instance 'future)))
     (when-resolved ,expression
       (lambda (result)
         (resolve-with-result hcfuture result))
       :catch ,@cases)
     hcfuture))

#|
; FUTURISED-OR
; Respect the sequential semantics of Common Lisp OR.
(defmacro futurised-or (&rest args)
  `(let ((fut (make-instance 'future)))
    (if (> (length ',args) 0)
        (process-or fut ,(car args) ',(cdr args))
      (resolve-with-result fut nil))
    fut))

(defun process-or (fut arg args)
  (if (cdr args)
      (when-resolved arg (lambda (or-result)
                                   (if or-result
                                       (resolve-with-result fut or-result)
                                     (process-or fut (car args) (cdr args)))))
     (when-resolved arg (lambda (or-result)
                                  (resolve-with-result fut or-result)))))
|#

; FUTURISED-OR
; Respect the sequential semantics of Common Lisp OR.
(defmacro futurised-or (&rest args)
  `(let ((fut (make-instance 'future)))
    (if (> ,(length args) 0)
        ,(process-or `fut args)
      (resolve-with-result fut nil))
    fut))

(defun process-or (fut args)
  (if (cdr args)
      `(when-resolved ,(car args) (lambda (or-result)
                                   (if or-result
                                       (resolve-with-result ,fut or-result)
                                     ,(process-or fut (cdr args)))))
     `(when-resolved ,(car args) (lambda (or-result)
                                  (resolve-with-result ,fut or-result)))))

; FUTURISED-AND
; Respect the sequential semantics of Common Lisp AND
(defmacro futurised-and (&rest args)
  `(let ((fut (make-instance 'future)))
     (if (> ,(length args) 0)
         ,(process-and `fut args)
       (resolve-with-result fut t))
     fut))

(defun process-and (fut args)
  (if (cdr args)
      `(when-resolved ,(car args)
                      (lambda (and-result)
                        (if and-result
                            ,(process-and fut (cdr args))
                          (resolve-with-result ,fut nil))))
    `(when-resolved ,(car args)
                    (lambda (and-result)
                      (resolve-with-result ,fut and-result)))))

; FUTURISE-LAMBDA
(defmacro futurised-lambda (&rest forms)
  (multiple-value-bind (lambda-list declarations
                             documentation body)
      (parse-lambda forms)
    `(lambda ,lambda-list       
       ,@(when documentation `(,documentation))
       (let ((lambda-future (make-instance 'future)))
         (when-all-resolved (list ,@lambda-list)
           (lambda (results)
             (let (,@(when lambda-list
                       (loop for i from 0 to (1- (length lambda-list))
                             collect `(,(nth i lambda-list) (nth ,i results)))))
               ,@(when declarations `(,@declarations))
               (resolve-with-result lambda-future
                                    (futurised-progn
                                     ,@body)))))
         (if (resolvedp lambda-future)
             (future-result lambda-future)
           lambda-future)))))

;;;;;;;;;;;;;
; FUNCTIONS ;
;;;;;;;;;;;;;

; Execute the continuation lambda when any of the
; futures of the list passed as parameter is resolved.
; As parameters, pass the future's result and a boolean
; indicating if all the futures are resolved.
(defun when-any-resolved (futures continuation-lambda)
  (let ((resolved-futures nil))
    (loop for future in futures
          do (when-resolved future
               (lambda (any-result)
                 (push any-result resolved-futures)
                 (funcall continuation-lambda any-result
                          (eql (list-length futures)
                               (list-length resolved-futures))))))))

; Execute the lambda only after the resolution of the
; first future.
(defun when-first-resolved (futures continuation-lambda)
  (let ((first-resolvedp nil))
    (when-any-resolved futures
      (lambda (first-result lastp)
        (declare (ignore lastp))
        (unless first-resolvedp
          (setf first-resolvedp t)
          (funcall continuation-lambda first-result))))))

; PARALLEL-AND
; Check the trueness of expressions in parallel.
; Return nil upon the first nil result.
; Otherwise return the last resolved (non-nil) result.  
(defun parallel-and (&rest args)
  (let ((pand-future (make-instance 'future)))
    (when-any-resolved args 
      (lambda (pand-result last-resultp)
        (unless (resolvedp pand-future)
          (if pand-result
              (when last-resultp
                (resolve-with-result pand-future pand-result))
            (resolve-with-result pand-future nil)))))
    pand-future))

; PARALLEL-OR
; Check the trueness of expression in parallel.
; Return the first non-nil resolved result Otherwise
; return nil.
(defun parallel-or (&rest args)
  (let ((por-future (make-instance 'future)))
    (when-any-resolved args 
      (lambda (por-result last-resultp)
        (unless (resolvedp por-future)
          (if por-result
              (resolve-with-result por-future por-result)
            (when last-resultp
              (resolve-with-result por-future nil))))))
    por-future))



; FUTURISED-REDUCE
; This is a limited version of the Common Lisp REDUCE.
; It does not include the keyword arguments: key, from-end,
; start, end, and initial-value.
(defun futurised-reduce (continuation-lambda futures)
  (let ((reduce-future (make-instance 'future)))
    (when-all-resolved futures
      (lambda (results)
        (resolve-with-result reduce-future
                             (funcall 'reduce continuation-lambda
                                      results))))
    reduce-future))

; FUTURISED-REMOVE-IF
; This is a limited version of the Common Lisp REMOVE-IF.
; It does not include the keyword arguments: from-end,
; start, end, count, and key.
(defun futurised-remove-if (test sequence)
  (when sequence
    (let ((remove-future (make-instance 'future))
          (futures (loop for element in sequence
                         collect (futurised-funcall test element))))
      (when-all-resolved futures
                         (lambda (results)
                           (resolve-with-result remove-future
                                                (loop for element in sequence
                                                      for result in results
                                                      unless result
                                                      collect element))))
      (if (resolvedp remove-future)
          (future-result remove-future)
        remove-future))))

; FUTURISED-FIND-IF
; This is a limited version of the Common Lisp FIND-IF.
; It does not include the keyword arguments: from-end,
; start, end, count, and key.
(defun futurised-find-if (test sequence)
  (when sequence
    (let ((find-future (make-instance 'future))
          (futures (loop for element in sequence
                         collect (futurised-funcall test element))))
      (when-all-resolved futures
                         (lambda (results)
                           (resolve-with-result find-future
                                                (loop for element in sequence
                                                      for result in results
                                                      when result
                                                      return element))))
      (if (resolvedp find-future)
          (future-result find-future)
        find-future))))

; FUTURISED-APPLY
(defun futurised-apply (future-function args)
  (let ((apply-future (make-instance 'future)))
    (when-resolved future-function
      (lambda (function)
        (resolve-with-result apply-future
                             (apply function args))))
    apply-future))

; FUTURISED-FUNCALL
(defun futurised-funcall (future-function &rest args)
  (futurised-apply future-function args))








      
           
