;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;  (__)  LAMBIC Project
;   )(   Written by Jorge Vallejos
;        
; (c) Vrije Universiteit Brussel, 2008 - 2010
;
; Permission is hereby granted, free of charge, to any person
; obtaining a copy of this software and associated documentation
; files (the "Software"), to deal in the Software without
; restriction, including without limitation the rights to use,
; copy, modify, merge, publish, distribute, sublicense, and/or
; sell copies of the Software, and to permit persons to whom the
; Software is furnished to do so, subject to the following
; conditions:
;
; The above copyright notice and this permission notice shall be
; included in all copies or substantial portions of the Software.
;
; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
; EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
; OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
; NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
; HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
; WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
; OTHER DEALINGS IN THE SOFTWARE.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; LAMBIC FUTURE-BASED SEQUENTIALITY
;
; MOP extension to enable methods: 
; 
; - to always return a future as result
; - to receive futures as parameters
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package :lambic)

(defvar *standard-timeout* 30)

;;;;;;;;;
; PROGN ;
;;;;;;;;;

; Extension to futures
; When when-resolved is called with a non-future value the continuation
; gets evaluated with that value.
(defmethod when-resolvedf (object continuation)
  (funcall continuation object))


; Futurised PROGN
; Execute the expressions contained in the progn sequentiallly. If during this
; execution an expression returns a future, wait for this future to be
; resolved before executing the rest of expressions.
; PROGN returns a future which is resolved when the Nth expression is executed.
(defmacro futurised-progn (&rest args)
  `(let ((fut (make-instance 'lambic:future)))
     ,(sequential-progn `fut args)
     (if (resolvedp fut)
         (future-result fut)
       fut)))

; INTERNAL USE ONLY
; Continuation of futurised PROGN.
(defun sequential-progn (fut args)
  (if (cdr args)
      `(when-resolved ,(car args)
                      (lambda (progn-result)
                        (declare (ignore progn-result))
                        ,(sequential-progn fut (cdr args))))
    `(when-resolved ,(car args)
                    (lambda (progn-result)
                      (resolve-with-result ,fut progn-result)))))


; PARALLEL-PROGN
; Execute the expressions contained in the progn in parallel. If the execution
; of a contained expression returns a future, do not wait for its resolution.
; PARALLEL-PROGN returns a future which is resolved when the Nth expression is
; executed.
(defmacro parallel-progn (&rest args)
  `(let ((fut (make-instance 'future)))
     (resolve-with-result fut (progn ,@args))
     fut))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; AUXILIARS FOR DEFINE-PEER-CLASS ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Remove accessors of slots.
; BIG ASSUMPTION: :accessor pair at the end of the slot
; definition.
(defun update-slots (slots)
  (loop for slot in slots
        if (position :accessor slot)
        collect (let* ((new-slot (remove :accessor slot)))
                  (butlast new-slot)) into updated-slots
        and 
        collect (cons (car slot)
                      (nth (1+ (position :accessor slot)) slot)) into accessor-conses
        else
        collect slot into updated-slots
        finally return (values updated-slots accessor-conses)))

; Ensure that accessor functions are instances of FUTURISED-FUNCTION class.
(defun ensure-futurised-accessors (accessor-conses class-name)
  (loop for accessor-cons in accessor-conses
        collect 
         `(progn
            (define-peer-method ,(cdr accessor-cons) ((object ,class-name))
              (slot-value object ',(car accessor-cons)))
            (define-peer-method (setf ,(cdr accessor-cons)) (value (object ,class-name))
              (setf (slot-value object ',(car accessor-cons)) value)))))


#|
; Ensure that accessor functions are instances of FUTURISED-FUNCTION class.
(defun ensure-futurised-accessors (accessor-conses)
  (loop for accessor-cons in accessor-conses
        collect 
         `(progn
            ;(define-peer-generic ,(cdr accessor-cons) (object))
            (define-peer-method ,(cdr accessor-cons) (object)
              (slot-value object ',(car accessor-cons)))
            ;(define-peer-generic (setf ,(cdr accessor-cons)) (value object))
            (define-peer-method (setf ,(cdr accessor-cons)) (value object)
              (setf (slot-value object ',(car accessor-cons)) value)))))
|#

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; FUTURISED FUNCTION EXECUTION ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; Unfold method arguments. For each argument future, replace the future
; with the future's result.
(defun unfold-all (arguments)
  (let ((unfolded-arguments nil))
    (loop for argument in arguments
          do (push (unfold argument) unfolded-arguments))
    (reverse unfolded-arguments)))

; Replace the argument future with the future's result.
(defun unfold (argument)
  (if (typep argument 'future)
      (future-result argument)
    argument))

; Execute the continuation lambda after all the futures passed
; as parameter are resolved. Pass the list of results of the
; futures as parameter to the lambda.
(defun when-all-resolved (futures continuation-lambda)
  (let ((multiple-future (make-instance 'future))
        (resolved-futures 0))
    (when-resolved multiple-future continuation-lambda)
    (loop for future in futures
          do (when-resolved future
               (lambda (useless-result)
                 (declare (ignore useless-result))
                 (incf resolved-futures)
                 (when (equal resolved-futures (list-length futures))
                   (resolve-with-result multiple-future
                                        (unfold-all futures))))))))

#|
(defmethod no-applicable-method ((function futurised-function) &rest args)
  (debug-print "NO-APPLICABLE-METHOD for" (generic-function-name function))
  (debug-print "WITH ARGS" args)
  (when (in-current-actor (car args))
    (call-next-method)))
|#

; Process compute-discriminating-function metamethod for futurised function
(defmethod compute-discriminating-function ((function futurised-function))
  (lambda (&rest args)
    (let* ((result-future (make-instance 'future))
           (process-arg (first args))
           (unresolved-futures (loop for argument in args
                                     when (and (typep argument 'future)
                                               (not (resolvedp argument)))
                                     collect argument)))
      (let ((discriminating-function (call-next-method)))
      ;(format *screen-output* "DEBUG. FUNCTION: ~d~%DEBUG. DISCRIMINATING-FUNCTION: ~d~%"
      ;        (generic-function-name function) discriminating-function)
      (if unresolved-futures
          (when-all-resolved unresolved-futures
                             (lambda (useless-result)
                               (declare (ignore useless-result))
                               (dispatch function args process-arg result-future
                                         discriminating-function)))
        (dispatch function args process-arg result-future discriminating-function))
      (if (resolvedp result-future)
          (future-result result-future)
        result-future)))))
;      result-future)))

; Invoke the original method. Make asynchronous remote invocation if
; the first argument is a FAR-REFERENCE. If result is a future 
(defun dispatch (function args process-arg result-future discriminating-function)
  (let* ((unfolded-args (unfold-all args))
         (unfolded-process-arg (unfold process-arg))
         (applicable-methods-future (compute-peer-applicable-methods function
                                                                     unfolded-args)))
    (when-resolved applicable-methods-future
      (lambda (applicable-methods)
        (let ((result (if applicable-methods
                          (apply discriminating-function unfolded-args)
                        (if (in-current-actor unfolded-process-arg)
                            (apply 'no-applicable-method (list function unfolded-args))
                          (send-to-actor-of unfolded-process-arg
                                            (generic-function-name function)
                                            unfolded-args
                                            :due-in *standard-timeout*)))))
          ; when cam & rpa -> apply
          ; when cam & ~rpa -> apply
          ; when ~cam & rpa -> send
          ; when ~cam & ~rpa -> error
          (resolve-with-result result-future result))))))

#|
; Process compute-discriminating-function metamethod for futurised function
(defmethod compute-discriminating-function ((function futurised-function))
  (lambda (&rest args)
    (let* ((result-future (make-instance 'future))
           (process-arg (first args))
           (unresolved-futures (loop for argument in args
                                     when (and (typep argument 'future)
                                               (not (resolvedp argument)))
                                     collect argument))
           (discriminating-function (and (in-current-actor process-arg)
                                         (call-next-method))))
      (if unresolved-futures
          (when-all-resolved unresolved-futures
                             (lambda (useless-result)
                               (declare (ignore useless-result))
                               (dispatch function args process-arg result-future
                                         discriminating-function)))
        (dispatch function args process-arg result-future discriminating-function))
      (if (resolvedp result-future)
          (future-result result-future)
        result-future))))
;      result-future)))

; Invoke the original method. Make asynchronous remote invocation if
; the first argument is a FAR-REFERENCE. If result is a future 
(defun dispatch (function args process-arg result-future 
                          &optional discriminating-function)
  (let* ((unfolded-args (unfold-all args))
         (unfolded-process-arg (unfold process-arg))
         (result (if (in-current-actor unfolded-process-arg)
                     (apply discriminating-function unfolded-args)
                   (send-to-actor-of unfolded-process-arg
                                     (generic-function-name function)
                                     unfolded-args
                                     :due-in *standard-timeout*))))
    (resolve-with-result result-future result)))
|#

;;;;;;;;;;;;;;;;;;;;;;;;;;;
; CACHE FUNCTIONS PATTERN ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Thus far this only works for accessor functions.

; where hastables' keys and values are a method-name
; and futures respectively.
(defvar *cache-provider-functions* (make-hash-table))

; where hastables' keys and values are a method-name
; and futures respectively.
(defvar *cache-observer-functions* (make-hash-table))

; Return cache function's future
(defun find-cache-function-future (remote-object &key observer-function 
                                                provider-function)
  (let* ((cache-pairs (gethash observer-function *cache-observer-functions*))
         (cache-future (cdr (find-if #'(lambda (pair)
                                         (eql (car pair) remote-object))
                                     cache-pairs))))
    (unless cache-future
      (setf cache-future (make-instance 'future))
      (register-cache-observer-function remote-object observer-function cache-future)
      (in-actor-of remote-object 
                   (register-cache-provider-function remote-object
                                                     (generic-function-name provider-function)
                                                     cache-future)))
    cache-future))

; Return list of cache function's (observer) futures
(defun find-cache-function-futures (local-object provider-function)
  (let ((cache-pairs (gethash provider-function *cache-provider-functions*)))
    ;(debug-print "cache-pairs" cache-pairs)
    ;(debug-print "in writter" (generic-function-name provider-function))
    (loop for pair in cache-pairs
          when (eql (car pair) local-object)
          collect (cdr pair))))

; Register a function as cache function observer
(defun register-cache-observer-function (remote-object observer-function cache-future)
  (setf (gethash observer-function *cache-observer-functions*)
        (append (gethash observer-function *cache-observer-functions*)
                (list (cons remote-object cache-future)))))

; Register a function as cache function provider
(defun register-cache-provider-function (remote-object provider-function-name cache-future)
  (let ((provider-function (fdefinition provider-function-name)))
    (setf (gethash provider-function *cache-provider-functions*)
          (append (gethash provider-function *cache-provider-functions*)
                  (list (cons remote-object cache-future))))))
  